# -*- coding: UTF-8 -*-

import os
from typing import Optional, Union, Tuple

import numpy as np
import pandas as pd

from matplotlib import pyplot as plt
from pandas import DataFrame

from .. import util as ul
from ..util import path, collection, plot_end

__name__: str = "plot_radar"


def radar(
    ax_x: collection,
    ax_y: collection,
    title: str = None,
    colors: collection = None,
    width: float = 4,
    height: float = 4,
    center_text: str = None,
    rotation: float = 25,
    value_top: float = 0.1,
    text_top: float = 1.2,
    is_fixed: bool = False,
    is_angle: bool = True,
    y_limit: Tuple = (-0.5, 1),
    y_axis_scale: Tuple = (0, 1),
    output: path = None,
    show: bool = True
):
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        fig, ax = plt.subplots(figsize=(width, height), subplot_kw=dict(polar=True))

        ax_x = list(ax_x)
        ax_y = list(ax_y)

        # Create a circular bar chart
        theta = np.linspace(0, 2 * np.pi, len(ax_x), endpoint=False).tolist()
        ax_y += ax_y[:1]
        theta += theta[:1]

        width = 2 * 2.7 / len(ax_x)

        bars = ax.bar(theta, ax_y, width=width, color=colors, edgecolor='none', alpha=0.8, zorder=3)

        # Add category labels
        ax.set_xticks(theta)
        ax.set_xticklabels([])

        # Set y-axis range
        ax.set_ylim(y_limit[0], y_limit[1])

        # Remove the scale value of the circle
        ax.set_yticks(np.linspace(y_axis_scale[0], y_axis_scale[1], 6))  # Set the y-axis scale position
        ax.set_yticklabels([])  # Do not display scale values
        ax.set_theta_zero_location('N')  # Set polar axis position
        ax.set_theta_direction(-1)  # The angle increases counterclockwise

        # Add numerical labels
        for i, bar in enumerate(bars):
            height = bar.get_height()
            angle = np.degrees(theta[i])
            label_position = theta[i]
            ax.text(label_position, height + value_top if not is_fixed else value_top, round(height, 3), ha='center', va='center', color='#1f1f1f', rotation=-angle + rotation if is_angle else rotation)

        # Add radar line
        ax.plot(theta, ax_y, color='gray', linewidth=1, zorder=1)
        # Draw radar map
        ax.fill(theta, ax_y, color='#DDDDDD', alpha=0.1, zorder=2)

        if center_text is not None:
            ax.text(0, y_limit[0], center_text, ha='center', va='center', fontsize=14, color='black', zorder=11)

        if title is not None:
            ax.set_title(title, va='bottom')

        # Set the y-axis scale line color to light gray
        ax.tick_params('y', colors='#DDDDDD', grid_alpha=0.6, zorder=8)

        # Set the color of the outermost circle line
        ax.spines['polar'].set_color('#DDDDDD')

        plt.grid(axis='x', linestyle='-', alpha=0.4, zorder=9)

        # Draw peripheral category labels
        for i, label in enumerate(ax_x):
            angle = np.degrees(theta[i])
            ax.text(theta[i], text_top, label, ha='center', va='center', color='#1f1f1f', zorder=20, rotation=-angle + rotation if is_angle else rotation)

        # Adjust the layout to prevent label overlap
        plt.tight_layout()

        plot_end(fig, output, show)


def radar_trait(
    trait_df: DataFrame,
    trait_name: str = "All",
    trait_column_name: str = "id",
    value: str = "rate",
    clusters: str = "clusters",
    color: Union[collection, str] = None,
    clusters_sort: Optional[list] = None,
    width: float = 4,
    height: float = 4,
    rotation: float = 65,
    title: str = None,
    value_top: float = 0.1,
    text_top: float = 1.2,
    is_fixed: bool = False,
    is_angle: bool = True,
    y_limit: Tuple = (-0.5, 1),
    y_axis_scale: Tuple = (0, 1),
    output: path = None,
    show: bool = True
):

    def trait_plot(trait_: str, cell_df_: DataFrame) -> None:
        """
        show plot
        :param trait_: trait name
        :param cell_df_:
        :return: None
        """
        ul.log(__name__).info("Plotting bar {}".format(trait_))
        trait_score = cell_df_[cell_df_[trait_column_name] == trait_]

        # Sort
        if clusters_sort is not None:
            trait_score[clusters] = pd.Categorical(trait_score[clusters], categories=clusters_sort, ordered=True)
            trait_score = trait_score.sort_values(by=clusters)
            ax_x = clusters_sort
        else:
            trait_score = trait_score.sort_values([value], ascending=False)
            ax_x = trait_score[clusters].tolist()

        colors = None
        if color is not None:
            if isinstance(color, str):
                if color in trait_score.columns:
                    colors = trait_score[color]
            elif isinstance(color, collection):
                colors = color

        radar(
            ax_x=ax_x,
            ax_y=trait_score[value].tolist(),
            title=f"{title} {trait_}" if title is not None else title,
            colors=colors,
            width=width,
            height=height,
            rotation=rotation,
            value_top=value_top,
            text_top=text_top,
            is_fixed=is_fixed,
            is_angle=is_angle,
            y_limit=y_limit,
            y_axis_scale=y_axis_scale,
            center_text=trait_,
            output=os.path.join(output, f"{trait_}_enrichment_radar.pdf") if output is not None else None,
            show=show
        )

    trait_list = list(set(trait_df[trait_column_name]))
    # judge trait
    if trait_name != "All" and trait_name not in trait_list:
        ul.log(__name__).error(
            f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}, "
            f"Suggest modifying the {trait_column_name} parameter information"
        )
        raise ValueError(
            f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}, "
            f"Suggest modifying the {trait_column_name} parameter information"
        )

    # plot
    if trait_name == "All":
        for trait in trait_list:
            trait_plot(trait_=trait, cell_df_=trait_df)
    else:
        trait_plot(trait_name, trait_df)



# -*- coding: UTF-8 -*-

import warnings
from typing import Optional, Union, Literal

import numpy as np
import pandas as pd
from tqdm import tqdm
from anndata import AnnData
from pandas import DataFrame

from .. import util as ul
from ..util import check_adata_get, to_dense

__name__: str = "preprocessing_gsea"

_Datasets = Optional[Literal['Human', 'Mouse', 'Yeast', 'Fly', 'Fish', 'Worm']]


def gsea_enrichr(
    gene_list: list[str],
    gene_sets: Union[list[str], set] = (
        "GO_Biological_Process_2023",
        "GO_Cellular_Component_2023",
        "GO_Molecular_Function_2023",
        "GWAS_Catalog_2023",
        "KEGG_2016"
    ),
    organism: _Datasets = "human",
    is_verbose: bool = True,
    output_dir: Optional[str] = None
) -> DataFrame:

    import gseapy as gp

    # noinspection PyTypeChecker
    names = gp.get_library_name(organism)

    if not set(gene_sets).issubset(set(names)):
        ul.log(__name__).error(f"The set of the {gene_sets} needs to include `gp.get_library_name(organism)`")
        raise ValueError(f"The set of the {gene_sets} needs to include `gp.get_library_name(organism)`")

    if output_dir is not None:
        ul.file_method(__name__).makedirs(output_dir)

    if is_verbose:
        ul.log(__name__).info("GSEA enrichr.")

    # noinspection PyTypeChecker
    gsea = gp.enrichr(gene_list=gene_list, gene_sets=list(gene_sets), organism=organism, outdir=output_dir)
    return gsea.results


def get_gene_enrichment(
    adata: AnnData,
    top_number: int = 50,
    threshold: float = 0.01,
    layer: Optional[str] = None,
    is_order_or_lt: bool = True,
    is_top: bool = True,
    gene_sets: Union[list[str], set] = (
        "GO_Biological_Process_2023",
        "GO_Cellular_Component_2023",
        "GO_Molecular_Function_2023",
        "GWAS_Catalog_2023",
        "KEGG_2016"
    ),
    organism: _Datasets = "human",
    output_dir: Optional[str] = None,
) -> DataFrame:
    ul.log(__name__).info("Gene enrichment analysis.")
    # get data
    new_adata = check_adata_get(adata, layer=layer)

    enrichr_data_list = []

    cluster_list: list[str] = list(new_adata.var.index)
    gene_list: list = list(new_adata.obs_names)

    if is_top and len(gene_list) < top_number:
        ul.log(__name__).warning(f"The number of parameters `top_number` ({top_number}) is greater than the number of genes ({len(gene_list)}).")
        top_number = len(gene_list)

    # Add data
    for cluster in tqdm(cluster_list):

        # get index
        _values_ = to_dense(new_adata[:, cluster].X, is_array=True).flatten()

        if is_top:
            _index_ = np.argsort(_values_)[0:top_number] if is_order_or_lt else np.argsort(_values_)[-top_number:]
            _cluster_gene_list_: list[str] = list(np.array(gene_list)[_index_])
        else:
            _index_ = _values_ <= threshold if is_order_or_lt else _values_ >= threshold
            _cluster_gene_list_: list[str] = list(np.array(gene_list)[_index_])

        if len(_cluster_gene_list_) == 0:
            ul.log(__name__).warning(f"The gene list for cluster {cluster} is empty.")
            continue

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            enrichr_data: DataFrame = gsea_enrichr(
                gene_list=_cluster_gene_list_,
                gene_sets=gene_sets,
                organism=organism,
                is_verbose=False,
                output_dir=output_dir
            )

        enrichr_data["cluster"] = cluster
        enrichr_data_list.append(enrichr_data)

        del _values_, _index_, _cluster_gene_list_

    data = pd.concat(enrichr_data_list, axis=0)

    return data



# -*- coding: UTF-8 -*-

from ._read_ import read_barcodes_file, read_sc_atac, read_sc_atac_10x_h5, read_h5ad, read_h5, read_variants, read_pkl
from ._write_ import to_meta, to_fragments, save_h5ad, save_h5, save_pkl

__all__ = [
    "read_barcodes_file",
    "read_sc_atac_10x_h5",
    "read_sc_atac",
    "read_variants",
    "read_h5ad",
    "read_pkl",
    "read_h5",
    "save_h5",
    "save_h5ad",
    "save_pkl",
    "to_meta",
    "to_fragments"
]



# -*- coding: UTF-8 -*-

from ._constant_ import (
    project_version,
    project_name,
    project_cache_path,
    is_form_log_file,
    log_file_path,
    path,
    sparse_array,
    sparse_matrix,
    sparse_data,
    dense_data,
    matrix_data,
    chrtype,
    number,
    collection,
    enrichment_optional,
    difference_peak_optional,
    plot_color_types,
    type_50_colors,
    plot_cmap_50,
    type_20_colors,
    plot_cmap_20,
    type_set_colors,
    plot_cmap_set
)

from ._core_ import (
    file_method,
    log,
    to_dense,
    to_sparse,
    sum_min_max,
    get_index,
    list_duplicate_set,
    split_matrix,
    merge_matrix,
    list_index,
    numerical_bisection_step,
    get_real_predict_label,
    strings_map_numbers,
    down_sampling_data,
    set_inf_value,
    plot_start,
    plot_end,
    generate_str,
    matrix_dot_block_storage,
    matrix_multiply_block_storage,
    sparse_matrix_operation_memory_efficient,
    vector_multiply_block_storage,
    matrix_division_block_storage,
    matrix_callback_block_storage,
    check_adata_get,
    add_cluster_info,
    generate_hex_colors,
    check_gpu_availability
)

__all__ = [
    "project_version",
    "project_name",
    "project_cache_path",
    "is_form_log_file",
    "log_file_path",
    "file_method",
    "log",
    "path",
    "plot_color_types",
    "sparse_array",
    "sparse_matrix",
    "sparse_data",
    "dense_data",
    "matrix_data",
    "number",
    "plot_cmap_20",
    "plot_cmap_50",
    "plot_cmap_set",
    "type_set_colors",
    "get_real_predict_label",
    "to_dense",
    "to_sparse",
    "list_duplicate_set",
    "chrtype",
    "type_50_colors",
    "type_20_colors",
    "sum_min_max",
    "numerical_bisection_step",
    "split_matrix",
    "merge_matrix",
    "collection",
    "enrichment_optional",
    "difference_peak_optional",
    "set_inf_value",
    "plot_start",
    "plot_end",
    "generate_str",
    "list_index",
    "strings_map_numbers",
    "add_cluster_info",
    "check_adata_get",
    "down_sampling_data",
    "get_index",
    "matrix_dot_block_storage",
    "matrix_multiply_block_storage",
    "sparse_matrix_operation_memory_efficient",
    "vector_multiply_block_storage",
    "matrix_division_block_storage",
    "matrix_callback_block_storage",
    "generate_hex_colors",
    "check_gpu_availability"
]



# -*- coding: UTF-8 -*-

import math
import os
import pickle
import random
import shutil
import string
from typing import Tuple, Union, Literal

import numpy as np
import pandas as pd
import torch
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from numpy import asarray
from anndata import AnnData
from pandas import DataFrame
from scipy import sparse
from tqdm import tqdm

from yzm_file import StaticMethod
from yzm_log import Logger

from .. import util as ul
from ._constant_ import dense_data, sparse_data, sparse_matrix, matrix_data, number, collection, project_name

__name__: str = "util_core"


def file_method(name: str = None) -> StaticMethod:
    name = f"{project_name}_{name}" if name is not None else project_name
    return StaticMethod(log_file=os.path.join(ul.log_file_path, name), is_form_log_file=ul.is_form_log_file)


def log(name: str = None) -> Logger:
    name = f"{project_name}_{name}" if name is not None else project_name
    return Logger(name, log_path=os.path.join(ul.log_file_path, name), is_form_file=ul.is_form_log_file)


def to_dense(sm: matrix_data, is_array: bool = False) -> dense_data:
    """
    Convert sparse matrix to dense matrix
    :param sm: sparse matrix
    :param is_array: True converted to array form, False natural output
    :return: dense matrix
    """

    if sm is None:
        log(__name__).warning("The input matrix (sm parameter) is not feasible")
        return np.asmatrix([])

    if isinstance(sm, sparse_data):
        return np.array(sm.todense()) if is_array else sm.todense()

    dense_sm = sm.todense() if sparse.issparse(sm) else sm
    return np.array(dense_sm) if is_array else dense_sm


def to_sparse(dm: dense_data, way_callback=sparse.csr_matrix, is_matrix: bool = True) -> sparse_matrix:
    """
    Convert dense matrix to sparse matrix
    :param dm: dense matrix
    :param way_callback: How to form sparse matrix
    :param is_matrix: True converted to matrix form, False natural output
    :return: sparse matrix
    """

    if dm is None:
        log(__name__).warning("The input matrix (dm parameter) is not feasible")
        return sparse.coo_matrix([])

    if isinstance(dm, dense_data):
        return way_callback(dm) if is_matrix else dm

    sparse_m = dm if sparse.issparse(dm) else way_callback(dm)
    return way_callback(sparse_m) if is_matrix else sparse_m


def sum_min_max(data: matrix_data, axis: int = 1) -> Tuple[number, number]:
    """
    Obtain the minimum/maximum sum of rows in the matrix
    :param data: matrix data
    :param axis: {0, 1} 1: row, 0: col
    :return: Minimum value of rows, maximum value of rows
    """
    rows, cols = data.shape

    if rows == 0 or cols == 0:
        return 0, 0

    rows_sum = list(np.array(data.sum(axis=axis)).flatten())
    return min(rows_sum), max(rows_sum)


def get_index(position: number, positions_list: list) -> Union[number, Tuple[number, number]]:
    """
    Search for position information. Similar to half search.
        If the position exists in the list, return the index.
        If it does not exist, return the index located between the two indexes
    :param position: position
    :param positions_list: position list
    :return: position index
    """
    # sort
    positions_list.sort()
    # search
    position_size: int = len(positions_list)
    left, right = 0, position_size - 1

    while left <= right:
        mid = (left + right) // 2

        if positions_list[mid] == position:
            return mid
        elif positions_list[mid] > position:
            right = mid - 1
        else:
            left = mid + 1

    return right, left


def list_duplicate_set(data: list) -> list:
    """
    Append numbering to duplicate information
    :param data: input data
    :return: Unique data with constant quantity
    """

    if len(data) == len(set(data)):
        return data

    new_data = []
    range_data = range(len(data))

    is_warn: bool = False

    for i in range_data:

        # judge duplicate
        if data[i] not in new_data:
            new_data.append(data[i])
        else:
            j: int = 2

            while True:

                if not isinstance(data[i], str) and not is_warn:
                    log(__name__).warning("Convert non string types to string types.")
                    is_warn = True

                # format new index
                elem: str = data[i] + "+" + str(j)

                # Add index
                if elem not in new_data:
                    new_data.append(elem)
                    break

                j += 1

    return new_data


def get_sub_data(data: collection, size: int) -> collection:
    # get information
    old_size = len(data)

    if size >= old_size:
        log(__name__).warning("The given size is greater than the original data size")
        return data

    old_data: list = data.copy()
    rate = size / old_size
    # add sub data
    new_data: list = []

    for i in range(size):
        new_data.append(old_data[math.floor(rate * i)])

    return new_data


def split_matrix(data: matrix_data, axis: Literal[0, 1] = 0, chunk_number: int = 1000) -> list:
    # get size
    new_data = to_dense(data, is_array=True)
    rows, cols = new_data.shape

    # get number
    total: int = rows if axis == 0 else cols

    # get split number
    split_number = total // chunk_number

    # Determine whether to divide equally
    tail_number = total % chunk_number
    if tail_number != 0:
        split_number += 1

    log(__name__).info(f"Divide the matrix into {split_number} parts")

    # Add data
    split_data_list = []
    for i in range(split_number):
        # set index
        start_index: int = i * chunk_number
        end_index: int = total if (i + 1) * chunk_number > total else (i + 1) * chunk_number
        split_data = new_data[start_index:end_index, :] if axis == 0 else new_data[:, start_index:end_index]
        split_data_list.append(split_data)

    return split_data_list


def merge_matrix(datas: list, axis: Literal[0, 1] = 0) -> list:
    # get size
    size = len(datas)
    range_size = range(size)

    # get row col
    constant: int = datas[0].shape[1] if axis == 0 else datas[0].shape[0]
    total: int = 0
    shapes: list = []

    # get chunk size
    for i in range_size:
        shape = datas[i].shape
        judge = shape[1] if axis == 0 else shape[0]

        # judge size
        if judge != constant:
            log(__name__).error("Inconsistent traits in the input dataset set.")
            raise ValueError("Inconsistent traits in the input dataset set.")

        total += shape[0] if axis == 0 else shape[1]
        shapes.append(shape)

    # format matrix
    matrix_shape = (total if axis == 0 else constant, constant if axis == 0 else total)
    log(__name__).info(f"Merge matrix {matrix_shape} shape")
    matrix: matrix_data = np.zeros(matrix_shape)

    # merge matrix
    col_record: int = 0
    for i in range_size:
        col_i = datas[i].shape[0] if axis == 0 else datas[i].shape[1]
        if axis == 0:
            matrix[col_record:col_record + col_i, :] = datas[i]
        else:
            matrix[:, col_record:col_record + col_i] = datas[i]
        col_record += col_i

    return matrix


def list_index(data: list) -> Tuple[list, collection]:
    info: list = []

    size: int = len(data)

    types: asarray = np.unique(data)

    for type_ in types:
        type_info: list = []
        for i in range(size):
            if type_ == data[i]:
                type_info.append(i)
        info.append(set(type_info))

    return info, types


def numerical_bisection_step(min_value: float, max_value: float, step_length: float) -> Tuple[collection, int]:
    if min_value > max_value:
        log(__name__).error(f"`min_value` ({min_value}) must be smaller than `max_value` ({max_value}).")
        raise ValueError(f"`min_value` ({min_value}) must be smaller than `max_value` ({max_value}).")

    number_list: list = []

    i = 0
    while min_value <= max_value:
        number_list.append(min_value)
        i += 1
        min_value += step_length

    return number_list, i


def get_real_predict_label(
    df: DataFrame,
    map_cluster: Union[str, collection],
    clusters: str = "clusters",
    value: str = "value"
) -> Tuple[DataFrame, int, list]:
    df_sort: DataFrame = df.sort_values([value], ascending=False)

    # Obtain the type of positive set clustering corresponding to the trait
    cluster_list: list = []
    if isinstance(map_cluster, str):
        cluster_list.append(map_cluster)
    else:
        cluster_list = list(map_cluster)

    # total label size
    total_size = df.shape[0]

    # true label size
    df_sort.insert(0, "true_label", 0)
    df_sort.loc[df_sort[df_sort[clusters].isin(cluster_list)].index, "true_label"] = 1

    # predict label size
    df_cluster = df[df[clusters].isin(cluster_list)].copy()
    df_cluster_size = df_cluster.shape[0]
    predict_label = list(np.ones(df_cluster_size))
    predict_label.extend(np.zeros(total_size - df_cluster_size))
    df_sort.insert(0, "predict_label", 0)
    df_sort.loc[:, "predict_label"] = predict_label
    df_sort["predict_label"] = df_sort["predict_label"].astype(int)

    return df_sort, df_cluster_size, cluster_list


def strings_map_numbers(str_list: list, start: int = 0) -> list:
    # Create an empty dictionary to store the mapping of strings to numerical values
    mapping = {}

    # Traverse the list and assign a unique numerical value to each string
    for i, item in enumerate(set(str_list), start=start):  # 使用set去除重复项，并从1开始编号
        mapping[item] = i

    # Use list derivation to convert strings to corresponding numerical values
    numeric_list = [mapping[item] for item in str_list]

    return numeric_list


def down_sampling_data(data: Union[matrix_data | collection], sample_number: int = 1000000) -> list:
    """
    down-sampling
    :param data: Data that requires down-sampling;
    :param sample_number: How many samples (values) were down-sampled.
    :return: Data after down-sampling.
    """

    if isinstance(data, collection):

        # Judge data size
        if data.size <= sample_number:
            return list(data)

        index = np.random.choice(range(data.size), sample_number, replace=False)

        return list(np.array(data)[index])

    elif isinstance(data, matrix_data):

        # Judge data size
        if data.shape[0] * data.shape[1] <= sample_number:
            return list(to_dense(data, is_array=True).flatten())

        data = to_dense(data, is_array=True)
        row_count = data.shape[0]
        col_count = data.shape[1]

        if row_count < 0:
            log(__name__).error("The number of rows of data must be greater than zero")
            raise ValueError("The number of rows of data must be greater than zero")

        log(__name__).info(
            f"Kernel density estimation plot down-sampling data from {row_count * col_count} to {sample_number}")

        # get count
        count = row_count * col_count
        iter_number: int = count // sample_number
        iter_sample_number: int = sample_number // iter_number
        iter_sample_number_final: int = sample_number % iter_number

        if iter_sample_number < 1:
            log(__name__).error("The sampling data is too small, increase the `sample_number` parameter value")
            raise ValueError("The sampling data is too small, increase the `sample_number` parameter value")

        log(__name__).info(f"Divide and conquer {iter_number} chunks")

        # Create index container
        return_data: list = []

        for i in range(iter_number + 1):

            if iter_number < 50:
                log(__name__).info(f"Start {i + 1}th chunk, {(i + 1) / iter_number * 100}%")
            elif iter_number >= 50 and i % 50 == 0:
                log(__name__).info(f"Start {i + 1}th chunk, {(i + 1) / iter_number * 100}%")

            # Determine if it is the last cycle
            end_count: int = count if i == iter_number else (i + 1) * sample_number

            if iter_sample_number_final == 0:
                index = np.random.choice(range(i * sample_number, end_count), iter_sample_number, replace=False)
            else:
                per_iter_sample_number: int = iter_sample_number_final if i == iter_number else iter_sample_number
                index = np.random.choice(range(i * sample_number, end_count), per_iter_sample_number, replace=False)

            # Add index
            for j in index:
                # row
                row_index = j // col_count
                # column
                col_index = j % col_count

                if row_index >= row_count:
                    log(__name__).error(f"index ({row_index}) out of range ({row_count})")
                    raise IndexError(f"index ({row_index}) out of range ({row_count})")

                if col_index >= col_count:
                    log(__name__).error(f"index ({col_index}) out of range ({col_count})")
                    raise IndexError(f"index ({col_index}) out of range ({col_count})")

                return_data.append(data[row_index, col_index])

        return return_data
    else:
        ul.log(__name__).error("The input data type is incorrect.")
        raise ValueError("The input data type is incorrect.")


def matrix_dot_block_storage(
    data1: matrix_data,
    data2: matrix_data,
    block_size: int = 10000,
    data: matrix_data = None
) -> matrix_data:
    """
    Perform Cartesian product of two matrices through block storage method.
    :param data1: Matrix 1
    :param data2: Matrix 2
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed.
    :param data: Return the placeholder variables of the result matrix.
        If there is value, it will reduce the consumption of memory space.
    :return: Cartesian product result
    """
    n, m = data1.shape
    p, q = data2.shape

    if m != p:
        log(__name__).error(f"Need to meet the principle of matrix multiplication. ({m} != {p})")
        raise ValueError(f"Need to meet the principle of matrix multiplication. ({m} != {p})")

    if block_size <= 0 or n <= block_size and m <= block_size and q <= block_size:
        return np.dot(data1, data2)

    n_range = range(0, n, block_size)
    q_range = range(0, q, block_size)
    m_range = range(0, m, block_size)

    _cache_path_ = os.path.join(ul.project_cache_path, generate_str(50))
    file_method(__name__).makedirs(_cache_path_)

    # Store block data
    log(__name__).info("[matrix_dot_block_storage]: Store block data...")
    total_steps = len(n_range) * len(q_range) * len(m_range)
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for j in q_range:
                for k in m_range:
                    i_max = min(i + block_size, n)
                    j_max = min(j + block_size, q)
                    k_max = min(k + block_size, m)

                    _matrix_ = np.dot(data1[i:i_max, k:k_max], data2[k:k_max, j:j_max])

                    with open(os.path.join(_cache_path_, str(i) + str(j) + str(k) + ".pkl"), 'wb') as f:
                        pickle.dump(_matrix_, f)

                    del _matrix_
                    pbar.update(1)

    del data1, data2

    if data is None or data.shape != (n, q):
        data = np.zeros((n, q))

    # Read data
    log(__name__).info("[matrix_dot_block_storage]: Read block data...")
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for j in q_range:
                for k in m_range:
                    i_max = min(i + block_size, n)
                    j_max = min(j + block_size, q)

                    with open(os.path.join(_cache_path_, str(i) + str(j) + str(k) + ".pkl"), 'rb') as f:
                        _matrix_ = pickle.load(f)
                        data[i:i_max, j:j_max] += _matrix_
                        del _matrix_

                    pbar.update(1)

    shutil.rmtree(_cache_path_)

    return data


def matrix_multiply_block_storage(
    data1: matrix_data,
    data2: matrix_data,
    block_size: int = 10000,
    data: matrix_data = None
) -> matrix_data:
    """
    Perform Hadamard product of two matrices through block storage method.
    :param data1: Matrix 1
    :param data2: Matrix 2
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed.
    :param data: Return the placeholder variables of the result matrix.
        If there is value, it will reduce the consumption of memory space.
    :return: Hadamard product result
    """
    n, m = data1.shape
    n1, m1 = data2.shape

    if n != n1 or m != m1:
        log(__name__).error(
            f"Need to satisfy the multiplication principle of Hadamard products in matrices. ({(n, m)} != {n1, m1})"
        )
        raise ValueError(
            f"Need to satisfy the multiplication principle of Hadamard products in matrices. ({(n, m)} != {n1, m1})"
        )

    if block_size <= 0 or n <= block_size and m <= block_size:
        return np.multiply(data1, data2)

    n_range = range(0, n, block_size)
    m_range = range(0, m, block_size)

    _cache_path_ = os.path.join(ul.project_cache_path, generate_str(50))
    file_method(__name__).makedirs(_cache_path_)

    # Store block data
    log(__name__).info("[matrix_multiply_block_storage]: Store block data...")
    total_steps = len(n_range) * len(m_range)
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                _matrix_ = np.multiply(data1[i:i_max, k:k_max], data2[i:i_max, k:k_max])

                with open(os.path.join(_cache_path_, str(i) + str(k) + ".pkl"), 'wb') as f:
                    pickle.dump(_matrix_, f)

                del _matrix_
                pbar.update(1)

    del data1, data2

    if data is None or data.shape != (n, m):
        data = np.zeros((n, m))

    # Read data
    log(__name__).info("[matrix_multiply_block_storage]: Read block data...")
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                with open(os.path.join(_cache_path_, str(i) + str(k) + ".pkl"), 'rb') as f:
                    _matrix_ = pickle.load(f)
                    data[i:i_max, k:k_max] += _matrix_
                    del _matrix_

                pbar.update(1)

    shutil.rmtree(_cache_path_)

    return data


def sparse_matrix_operation_memory_efficient(
    data1: matrix_data,
    data2: matrix_data,
    chunk_size: int = 10000,
    default: float = 1e+8,
    operation: Literal['+', '-', '*', '/'] = '*'
) -> sparse_matrix:
    """
    Perform element-wise addition, subtraction, multiplication, and division on two sparse matrices by blocks, supporting memory-efficient processing.
    :param data1: Sparse matrix 1
    :param data2: Sparse matrix 2
    :param chunk_size: The size of the segmentation stored in block wise element-wise operation.
        If the value is less than or equal to zero, no block operation will be performed.
    :param default: 1e+8
    :param operation: Element-wise operation type, optional '+', '-', '*', '/'
    :return: Result sparse matrix (CSR format)
    """
    n, m = data1.shape
    n1, m1 = data2.shape

    # Element-wise operations follow below
    if n != n1 or m != m1:
        log(__name__).error(f"Need to satisfy the element-wise operation principle in matrices. ({(n, m)} != {n1, m1})")
        raise ValueError(f"Need to satisfy the element-wise operation principle in matrices. ({(n, m)} != {n1, m1})")

    data1 = to_sparse(data1)
    data2 = to_sparse(data2)

    result = sparse.lil_matrix(data1.shape)

    for i in range(0, data1.shape[0], chunk_size):
        i_end = min(i + chunk_size, data1.shape[0])

        chunk1 = data1[i:i_end, :]
        chunk2 = data2[i:i_end, :]

        if operation == '+':
            result_chunk = chunk1 + chunk2
        elif operation == '-':
            result_chunk = chunk1 - chunk2
        elif operation == '*':
            result_chunk = chunk1.multiply(chunk2)
        elif operation == '/':
            # Sparse matrix division: convert to dense, perform element-wise division, then convert back to sparse
            dense_chunk1 = chunk1.todense()
            dense_chunk2 = chunk2.todense()
            # Avoid division by zero
            dense_chunk2[dense_chunk2 == 0] = default
            result_chunk = dense_chunk1 / dense_chunk2
            result_chunk = sparse.csr_matrix(result_chunk)
        else:
            log(__name__).error(f"Unsupported operation: {operation}")
            raise ValueError(f"Unsupported operation: {operation}")

        result[i:i_end, :] = result_chunk

    return result.tocsr()


def vector_multiply_block_storage(
    data1: collection,
    data2: collection,
    block_size: int = 10000,
    data: matrix_data = None
) -> matrix_data:
    """
    Two vectors are broadcast in rows and columns respectively and multiplied by Hadamard product
    :param data1: Vector 1
    :param data2: Vector 2
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed.
    :param data: Return the placeholder variables of the result matrix.
        If there is value, it will reduce the consumption of memory space.
    :return: Matrix
    """

    vector1 = to_dense(data1, is_array=True).flatten()[:, np.newaxis]
    vector2 = to_dense(data2, is_array=True).flatten()

    del data1, data2

    n, _ = vector1.shape
    q = vector2.shape[0]

    if block_size <= 0 or n <= block_size and q <= block_size:
        return vector1 * vector2

    n_range = range(0, n, block_size)
    q_range = range(0, q, block_size)

    _cache_path_ = os.path.join(ul.project_cache_path, generate_str(50))
    file_method(__name__).makedirs(_cache_path_)

    # Store block data
    log(__name__).info("[vector_multiply_block_storage]: Store block data...")

    total_steps = len(n_range) * len(q_range)
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for j in q_range:
                i_max = min(i + block_size, n)
                j_max = min(j + block_size, q)

                _matrix_ = vector1[i:i_max, :] * vector2[j:j_max]

                with open(os.path.join(_cache_path_, str(i) + str(j) + ".pkl"), 'wb') as f:
                    pickle.dump(_matrix_, f)

                del _matrix_
                pbar.update(1)

    del vector1, vector2

    if data is None or data.shape != (n, q):
        data = np.zeros((n, q))

    # Read data
    log(__name__).info("[vector_multiply_block_storage]: Read block data...")
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for j in q_range:
                i_max = min(i + block_size, n)
                j_max = min(j + block_size, q)

                with open(os.path.join(_cache_path_, str(i) + str(j) + ".pkl"), 'rb') as f:
                    _matrix_ = pickle.load(f)
                    data[i:i_max, j:j_max] += _matrix_
                    del _matrix_

                pbar.update(1)

    shutil.rmtree(_cache_path_)

    return data


def matrix_division_block_storage(
    matrix: matrix_data,
    value: Union[float, int, collection, matrix_data],
    block_size: int = 10000,
    data: matrix_data = None
) -> matrix_data:
    """
    Dividing a matrix by another value, vector, or matrix.
    :param matrix: Matrix
    :param value: Value, vector, or matrix
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed.
    :param data: Return the placeholder variables of the result matrix.
        If there is value, it will reduce the consumption of memory space.
    :return: Matrix
    """

    n, m = matrix.shape

    is_status: int = -1

    if isinstance(value, collection):
        value = to_dense(value, is_array=True)
        q = value.shape[0]

        if q != n and q != m:
            log(__name__).error(f"Inconsistent dimensions cannot be divided. {(n, m)} != {q}")
            raise ValueError(f"Inconsistent dimensions cannot be divided. {(n, m)} != {q}")

        is_status = 1 if n == q else 0
    elif isinstance(value, matrix_data):
        is_status = 2

    if block_size <= 0 or n <= block_size and m <= block_size:
        return matrix / value

    n_range = range(0, n, block_size)
    m_range = range(0, m, block_size)

    _cache_path_ = os.path.join(ul.project_cache_path, generate_str(50))
    file_method(__name__).makedirs(_cache_path_)

    # Store block data
    log(__name__).info("[matrix_division_block_storage]: Store block data...")
    total_steps = len(n_range) * len(m_range)
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                if is_status == -1:
                    _matrix_ = matrix[i:i_max, k:k_max] / value
                elif is_status == 0:
                    _matrix_ = matrix[i:i_max, k:k_max] / value[k:k_max]
                elif is_status == 1:
                    _matrix_ = matrix[i:i_max, k:k_max] / value[i:i_max]
                elif is_status == 2:
                    _matrix_ = matrix[i:i_max, k:k_max] / value[i:i_max, k:k_max]
                else:
                    raise RuntimeError("nothingness")

                with open(os.path.join(_cache_path_, str(i) + str(k) + ".pkl"), 'wb') as f:
                    pickle.dump(_matrix_, f)

                del _matrix_
                pbar.update(1)

    del matrix, value

    if data is None or data.shape != (n, m):
        data = np.zeros((n, m))

    # Read data
    log(__name__).info("[matrix_division_block_storage]: Read block data...")
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                with open(os.path.join(_cache_path_, str(i) + str(k) + ".pkl"), 'rb') as f:
                    _matrix_ = pickle.load(f)
                    data[i:i_max, k:k_max] += _matrix_
                    del _matrix_

                pbar.update(1)

    shutil.rmtree(_cache_path_)

    return data


def matrix_callback_block_storage(
    matrix: matrix_data,
    callback,
    block_size: int = 10000,
    data: matrix_data = None
) -> matrix_data:
    """
    Dividing a matrix by another value, vector, or matrix.
    :param matrix: Matrix
    :param callback: callback function
    :param block_size: The size of the segmentation stored in block wise matrix multiplication.
        If the value is less than or equal to zero, no block operation will be performed.
    :param data: Return the placeholder variables of the result matrix.
        If there is value, it will reduce the consumption of memory space.
    :return: Matrix
    """

    n, m = matrix.shape

    n_range = range(0, n, block_size)
    m_range = range(0, m, block_size)

    if block_size <= 0 or n <= block_size and m <= block_size:
        return callback(matrix)

    _cache_path_ = os.path.join(ul.project_cache_path, generate_str(50))
    file_method(__name__).makedirs(_cache_path_)

    # Store block data
    log(__name__).info("[matrix_callback_block_storage]: Store block data...")
    total_steps = len(n_range) * len(m_range)
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                _matrix_ = callback(matrix[i:i_max, k:k_max])

                with open(os.path.join(_cache_path_, str(i) + str(k) + ".pkl"), 'wb') as f:
                    pickle.dump(_matrix_, f)

                del _matrix_
                pbar.update(1)

    del matrix

    if data is None or data.shape != (n, m):
        data = np.zeros((n, m))

    # Read data
    log(__name__).info("[matrix_callback_block_storage]: Read block data...")
    with tqdm(total=total_steps) as pbar:
        for i in n_range:
            for k in m_range:
                i_max = min(i + block_size, n)
                k_max = min(k + block_size, m)

                with open(os.path.join(_cache_path_, str(i) + str(k) + ".pkl"), 'rb') as f:
                    _matrix_ = pickle.load(f)
                    data[i:i_max, k:k_max] += _matrix_
                    del _matrix_

                pbar.update(1)

    shutil.rmtree(_cache_path_)

    return data


def generate_str(length: int = 10) -> str:
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(length))


def set_inf_value(matrix: matrix_data) -> None:
    # solve -Inf/Inf value
    matrix_inf = np.logical_and(np.isinf(matrix), matrix > 0)
    matrix__inf = np.logical_and(np.isinf(matrix), matrix < 0)

    # set inf
    if np.any(matrix_inf):
        matrix[matrix_inf] = np.max(matrix[~matrix_inf]) * 2

    # set -inf
    if np.any(matrix__inf):
        matrix[matrix__inf] = np.min(matrix[~matrix__inf]) / 2


def check_adata_get(adata: AnnData, layer: str = None, is_dense: bool = True, is_matrix: bool = False) -> AnnData:
    # judge input data
    if adata.shape[0] == 0:
        log(__name__).warning("Input data is empty")
        raise ValueError("Input data is empty")

    # get data
    data: AnnData = adata.copy()

    # judge layers
    if layer is not None:
        if layer not in list(data.layers):
            log(__name__).error("The `layer` parameter needs to include in `adata.layers`")
            raise ValueError("The `layer` parameter needs to include in `adata.layers`")

        data.X = to_dense(data.layers[layer], is_array=True) if is_dense \
            else to_sparse(data.layers[layer], is_matrix=is_matrix)
    else:
        data.X = to_dense(data.X, is_array=True) if is_dense else to_sparse(data.X, is_matrix=is_matrix)

    return data


def add_cluster_info(data: DataFrame, data_ref: DataFrame, cluster: str) -> DataFrame:
    new_data: DataFrame = data.copy()
    if data_ref is not None and cluster not in new_data.columns:

        new_data: DataFrame = pd.merge(new_data, data_ref, how="left", left_index=True, right_index=True)

        if "barcode_x" in new_data.columns:
            new_data["barcode"] = new_data["barcode_x"]
            new_data.drop("barcode_x", axis=1, inplace=True)

            if "barcode_y" in new_data.columns:
                new_data.drop("barcode_y", axis=1, inplace=True)

        if "barcodes_x" in new_data.columns:
            new_data["barcodes"] = new_data["barcodes_x"]
            new_data.drop("barcodes_x", axis=1, inplace=True)

            if "barcodes_y" in new_data.columns:
                new_data.drop("barcodes_y", axis=1, inplace=True)

    if cluster not in new_data.columns:
        log(__name__).error(f"`{cluster}` is not in `adata.obs.columns`.")
        raise ValueError(f"`{cluster}` is not in `columns` ({new_data.columns}).")

    return new_data


def check_gpu_availability(verbose: bool = True) -> bool:
    available = torch.cuda.is_available()

    if verbose:

        if available:
            log(__name__).info("GPU is available.")
            log(__name__).info(f"Number of GPUs: {torch.cuda.device_count()}")
            log(__name__).info(f"GPU Name: {torch.cuda.get_device_name(torch.cuda.current_device())}")
        else:
            log(__name__).info("GPU is not available.")

    return available


def plot_start(
    title: str = None,
    x_name: str = None,
    y_name: str = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0,
    output: str = None,
    show: bool = True,
):
    if output is None and not show:
        ul.log(__name__).error(f"At least one of the `output` and `show` parameters is required")
        raise ValueError(f"At least one of the `output` and `show` parameters is required")

    fig, ax = plt.subplots(figsize=(width, height))
    fig.subplots_adjust(bottom=bottom)

    if title is not None:
        plt.title(title)

    if x_name is not None:
        plt.xlabel(x_name)

    if y_name is not None:
        plt.xlabel(y_name, rotation=90)

    return fig, ax


def plot_end(fig, output: str, show: bool, dpi: float = 300):
    if output is not None:

        if output.endswith(".pdf"):

            with PdfPages(output) as pdf:
                pdf.savefig(fig)

        elif output.endswith(".png") or output.endswith(".jpg"):
            plt.savefig(output, dpi=dpi)
        else:
            plt.savefig(f"{output}.png", dpi=dpi)

    if show:
        plt.show()

    plt.close()


def generate_hex_colors(num_colors):
    colors = []

    while len(colors) < num_colors:
        color = "#{:02X}{:02X}{:02X}".format(
            random.randint(0, 255),
            random.randint(0, 255),
            random.randint(0, 255)
        )
        colors.append(color)

    return colors



# -*- coding: UTF-8 -*-

import os

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import ListedColormap
from matplotlib.gridspec import GridSpec
from pandas import DataFrame

from .. import util as ul
from ..util import path, type_50_colors, type_20_colors, plot_end

__name__: str = "plot_barcode"


def barcode_base(
    df: DataFrame,
    cluster_list: list,
    sort_column: str = "value",
    column: str = "clusters",
    width: float = 1,
    height: float = 3,
    trait_column_name: str = "id",
    title: str = None,
    cmap: str = "Oranges",
    bar_label: str = "TRS",
    is_ticks: bool = True,
    colors: list = None,
    ground_true: list = None,
    output: path = None,
    show: bool = True
):
    # sort
    df_sort = df.sort_values([trait_column_name, sort_column], ascending=False)

    # set index
    class_list = list(set(df_sort[column]))
    id_list = list(set(df_sort[trait_column_name]))
    df_sort["class_index"] = np.zeros(df_sort.shape[0])

    if colors is None:
        colors = type_20_colors if len(class_list) <= 20 else type_50_colors

    for i in class_list:

        if ground_true is not None:
            ground_true: list
            df_sort.loc[df_sort[df_sort[column] == i].index, ["class_index"]] = ground_true.count(i)
        else:
            df_sort.loc[df_sort[df_sort[column] == i].index, ["class_index"]] = cluster_list.index(i)

    class_index = np.array(df_sort["class_index"])

    # figure
    fig = plt.figure(figsize=(width, height))
    fig.subplots_adjust(left=0.03, right=0.97, top=0.99, bottom=0.01)

    plt.axis("off")

    if title is not None:
        plt.title(title)

    gs = GridSpec(20, 20)
    ax1 = fig.add_subplot(gs[:17, 11:14] if is_ticks else gs[:18, 11:14])
    ax2 = fig.add_subplot(gs[:17, :8] if is_ticks else gs[:18, :8])

    # span the whole figure
    ax1.set_axis_off()
    ax1.imshow(
        np.array(class_index).reshape(int(df_sort.shape[0] / len(id_list)), -1),
        cmap=ListedColormap(colors),
        aspect='auto',
        interpolation='nearest'
    )
    # ax1.tick_params(axis='x', rotation=90)

    ax2.set_axis_off()
    im2 = ax2.imshow(
        np.array(df_sort[sort_column]).reshape(int(df_sort.shape[0] / len(id_list)), -1),
        cmap=cmap,
        aspect='auto',
        interpolation='nearest'
    )

    # [left, bottom, width, height]
    cax = fig.add_axes([0.12, 0.09, 0.5, 0.04] if is_ticks else [0.1, 0.04, 0.5, 0.04])
    color_bar = plt.colorbar(im2, ax=ax2, cax=cax, label=bar_label, orientation='horizontal')
    color_bar.set_label(bar_label)

    ticks = np.linspace(round(df_sort[sort_column].min(), 2), round(df_sort[sort_column].max() - 0.05, 2), 3)
    color_bar.set_ticks(ticks if is_ticks else [])

    plot_end(fig, output, show)


def barcode_trait(
    trait_df: DataFrame,
    trait_name: str = "All",
    trait_column_name: str = "id",
    sort_column: str = "value",
    clusters: str = "clusters",
    cmap: str = "viridis",
    width: float = 1,
    height: float = 3,
    is_ticks: bool = True,
    colors: list = None,
    ground_true: list = None,
    title: str = None,
    output: path = None,
    show: bool = True
):
    data: DataFrame = trait_df.copy()
    cluster_list = list(set(trait_df[clusters]))

    def trait_plot(trait_: str, atac_cell_df_: DataFrame) -> None:
        """
        show plot
        :param trait_: trait name
        :param atac_cell_df_:
        :return: None
        """
        ul.log(__name__).info("Plotting barcode {}".format(trait_))
        # get gene score
        trait_score = atac_cell_df_[atac_cell_df_[trait_column_name] == trait_]
        # Sort gene scores from small to large
        barcode_base(
            df=trait_score,
            cluster_list=cluster_list,
            sort_column=sort_column,
            trait_column_name=trait_column_name,
            column=clusters,
            width=width,
            height=height,
            is_ticks=is_ticks,
            colors=colors,
            cmap=cmap,
            ground_true=ground_true,
            title=f"{title} {trait_}" if title is not None else title,
            output=os.path.join(output, f"cell_{trait_}_score_rank.pdf") if output is not None else None,
            show=show
        )

    # noinspection DuplicatedCode
    trait_list = list(set(data[trait_column_name]))
    # judge trait
    if trait_name != "All" and trait_name not in trait_list:
        ul.log(__name__).error(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")
        raise ValueError(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")

    # plot
    if trait_name == "All":
        for trait in trait_list:
            trait_plot(trait, trait_df)
    else:
        trait_plot(trait_name, trait_df)



# -*- coding: UTF-8 -*-

import warnings
from typing import Optional, Literal

import numpy as np
from anndata import AnnData
from pandas import DataFrame

from .. import util as ul
from ..file import save_h5ad
from ..util import add_cluster_info, matrix_data, set_inf_value

__name__: str = "preprocessing_scanpy"

_Method = Optional[Literal['logreg', 't-test', 'wilcoxon', 't-test_overestim_var']]


def filter_data(
    adata: AnnData,
    min_cells: int = 1,
    min_peaks: int = 1,
    min_peaks_counts: int = 1,
    min_cells_counts: int = 1,
    cell_rate: Optional[float] = None,
    peak_rate: Optional[float] = None,
    is_copy: bool = False,
    is_min_cell: bool = True,
    is_min_peak: bool = False
) -> AnnData:
    """
    Filter scATAC-seq data
    :param adata: scATAC-seq data
    :param min_peaks_counts: Minimum number of counts required for a peak to pass filtering
    :param min_cells: Minimum number of cells expressed required for a peak to pass filtering
    :param min_cells_counts: Minimum number of counts required for a cell to pass filtering
    :param min_peaks: Minimum number of peaks expressed required for a cell to pass filtering
    :param cell_rate: Removing the percentage of cell count in total cell count only takes effect when the min_cells
        parameter is None
    :param peak_rate: Removing the percentage of peak count in total peak count only takes effect when the min_peaks
        parameter is None
    :param is_copy: Do you want to deeply copy data
    :param is_min_cell: Whether to screen cells
    :param is_min_peak: Whether to screen peaks
    :return: scATAC-seq data
    """
    import scanpy as sc

    ul.log(__name__).info("Filter scATAC-seq data")

    if adata.shape[0] == 0:
        ul.log(__name__).error("The scATAC data is empty")
        raise ValueError("The scATAC data is empty")

    filter_adata = adata.copy() if is_copy else adata
    cells_count, peaks_count = filter_adata.shape

    if cell_rate is not None:

        if cell_rate <= 0 or cell_rate >= 1:
            ul.log(__name__).error("The parameter of `cell_rate` should be between 0 and 1.")
            raise ValueError("The parameter of `cell_rate` should be between 0 and 1.")

        _min_cells_ = int(filter_adata.shape[0] * cell_rate)

        if _min_cells_ > 1:
            min_cells = _min_cells_

    if peak_rate is not None:

        if peak_rate <= 0 or peak_rate >= 1:
            ul.log(__name__).error("The parameter of `peak_rate` should be between 0 and 1.")
            raise ValueError("The parameter of `peak_rate` should be between 0 and 1.")

        _min_peaks_ = int(filter_adata.shape[1] * peak_rate)

        if _min_peaks_ > 1:
            min_peaks = _min_peaks_

    ul.log(__name__).info(f"min cells: {min_cells}, min peaks: {min_peaks}")
    sc.pp.filter_genes(filter_adata, min_cells=min_cells)
    sc.pp.filter_cells(filter_adata, min_genes=min_peaks)

    # filter peaks and cell
    if is_min_peak:
        sc.pp.filter_genes(filter_adata, min_counts=min_peaks_counts)

    if is_min_cell:
        sc.pp.filter_cells(filter_adata, min_counts=min_cells_counts)
    else:
        sc.pp.filter_cells(filter_adata)

    # judge cells count
    if filter_adata.shape[0] == 0:
        ul.log(__name__).error(
            "After screening, the number of cells was 0. Suggest setting the `is_min_peak` parameter to `False` or "
            "lowering the `cell_rate` and `peak_rate` parameters to try again"
        )
        return filter_adata

    ul.log(__name__).info(
        f"Filtered out cells {cells_count - filter_adata.shape[0]}, "
        f"Filtered out peaks {peaks_count - filter_adata.shape[1]}"
    )
    ul.log(__name__).info(f"Size of filtered scATAC data: {filter_adata.shape}")
    filter_adata.uns["step"] = 0
    return filter_adata


def get_difference_genes(
    adata: AnnData,
    cluster: str,
    method: _Method = "wilcoxon",
    cell_anno: Optional[DataFrame] = None,
    diff_genes_file: Optional[str] = None
) -> AnnData:

    import scanpy as sc

    # add cell annotation information
    adata.obs = add_cluster_info(adata.obs, cell_anno, cluster)

    if "log1p" not in adata.uns_keys():
        ul.log(__name__).info("The `log1p` not detected in `adata.uns_keys`, `log1p` operation needs to be performed.")
        raise ValueError("The `log1p` not detected in `adata.uns_keys`, `log1p` operation needs to be performed.")

    if "base" not in adata.uns["log1p"].keys():
        adata.uns["log1p"].update({"base": None})

    # gene
    ul.log(__name__).info("Rank genes for characterizing groups.")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        sc.tl.rank_genes_groups(adata=adata, groupby=cluster, method=method, use_raw=False)

    # get difference genes for each `cluster`
    diff_genes = adata.uns['rank_genes_groups']['names']

    # gene names
    gene_list: list = list(adata.var.index)
    gene_list.sort()
    gene_dict: dict = dict(zip(gene_list, range(len(gene_list))))

    # obs
    cluster_info: DataFrame = adata.obs.copy().groupby([cluster], as_index=False).size()
    cluster_info.index = cluster_info[cluster].astype(str)

    cluster_list: list = cluster_info.index.tolist()
    cluster_list.sort()

    _shape_ = (adata.shape[1], cluster_info.shape[0])
    diff_genes_score_matrix: matrix_data = np.zeros(_shape_)
    diff_genes_p_value_matrix: matrix_data = np.zeros(_shape_)
    diff_genes_adjusted_p_value_matrix: matrix_data = np.zeros(_shape_)
    diff_genes_log2_fold_change_matrix: matrix_data = np.zeros(_shape_)
    del _shape_

    # cluster
    for _cluster_ in cluster_list:
        ul.log(__name__).info(f"Obtaining differentially expressed genes for `cluster` ({_cluster_}).")
        # obtain cluster difference gene data
        _cluster_data_: DataFrame = sc.get.rank_genes_groups_df(adata, group=_cluster_)
        _cluster_index_: int = cluster_list.index(_cluster_)

        # Add data value
        for _gene_name_, _score_, _p_value_, _adjusted_p_value_, _log2_fold_change_ in zip(
            _cluster_data_["names"],
            _cluster_data_["scores"],
            _cluster_data_["pvals"],
            _cluster_data_["pvals_adj"],
            _cluster_data_["logfoldchanges"]
        ):
            _gene_index_: int = gene_dict[_gene_name_]
            diff_genes_score_matrix[_gene_index_, _cluster_index_] = 0 if np.isnan(_score_) else _score_
            diff_genes_p_value_matrix[_gene_index_, _cluster_index_] = 1 if np.isnan(_p_value_) else _p_value_
            diff_genes_adjusted_p_value_matrix[_gene_index_, _cluster_index_] = 1 if np.isnan(_adjusted_p_value_) else _adjusted_p_value_
            diff_genes_log2_fold_change_matrix[_gene_index_, _cluster_index_] = 0 if np.isnan(_log2_fold_change_) else _log2_fold_change_

        del _cluster_data_, _cluster_index_

    set_inf_value(diff_genes_score_matrix)
    set_inf_value(diff_genes_p_value_matrix)
    set_inf_value(diff_genes_adjusted_p_value_matrix)
    set_inf_value(diff_genes_log2_fold_change_matrix)

    diff_genes_p_value_matrix[diff_genes_p_value_matrix == 0] = np.min(diff_genes_p_value_matrix[diff_genes_p_value_matrix != 0])
    diff_genes_adjusted_p_value_matrix[diff_genes_adjusted_p_value_matrix == 0] = np.min(diff_genes_adjusted_p_value_matrix[diff_genes_adjusted_p_value_matrix != 0])

    # create
    diff_genes_adata: AnnData = AnnData(diff_genes_score_matrix, obs=adata.var, var=cluster_info)
    diff_genes_adata.layers["p_value"] = diff_genes_p_value_matrix
    diff_genes_adata.layers["adjusted_p_value"] = diff_genes_adjusted_p_value_matrix
    diff_genes_adata.layers["log2_fold_change"] = diff_genes_log2_fold_change_matrix

    # Add diff_genes
    diff_genes_adata.uns["diff_genes"] = diff_genes

    if diff_genes_file is not None:
        save_h5ad(diff_genes_adata, diff_genes_file)

    return diff_genes_adata



# -*- coding: UTF-8 -*-

import os
from typing import Tuple, Union

from pandas import DataFrame
from matplotlib import pyplot as plt
import seaborn as sns

from .. import util as ul
from ..util import path, plot_end

__name__: str = "plot_box"


def box_base(
    df: DataFrame,
    value: str = "value",
    x_name: str = None,
    y_name: str = "value",
    clusters: str = "clusters",
    palette: Union[Tuple, list] = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0.3,
    line_width: float = 0.3,
    marker_size: float = 0.2,
    rotation: float = 65,
    orient: str = None,
    title: str = None,
    whis: float = 1.5,
    show_fliers: bool = True,
    is_sort: bool = True,
    order_names: list = None,
    output: path = None,
    show: bool = True
) -> None:
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        # judge
        # noinspection DuplicatedCode
        df_columns = list(df.columns)

        if value not in df_columns:
            ul.log(__name__).error(f"The `value` ({value}) parameter must be in the `df` parameter data column name ({df_columns})")
            raise ValueError(f"The `value` ({value}) parameter must be in the `df` parameter data column name ({df_columns})")

        fig, ax = plt.subplots(figsize=(width, height))
        # noinspection DuplicatedCode
        fig.subplots_adjust(bottom=bottom)

        if title is not None:
            plt.title(title)

        group_columns = [clusters]

        new_df: DataFrame = df.groupby(group_columns, as_index=False)[value].median()

        if "color" in df_columns:
            new_df_color: DataFrame = df.groupby(group_columns, as_index=False)["color"].first()

            new_df = new_df.merge(new_df_color, how="left", on=clusters)

        colors: list = []

        # sort
        if is_sort:
            new_df.sort_values([value], ascending=False, inplace=True)
            y_names: Union[list, None] = list(new_df[clusters])
            if "color" in df_columns:
                colors = new_df["color"]
        else:
            new_df.index = new_df[clusters]
            if order_names is not None:
                y_names: list = order_names
                if "color" in df_columns:
                    for i in order_names:
                        for j, c in zip(new_df[clusters], new_df["color"]):
                            if i == j:
                                colors.append(c)
                                break
            else:
                y_names = new_df[clusters]
                if "color" in df_columns:
                    colors = new_df["color"]

        # scatter
        ax1 = sns.boxplot(
            data=df,
            x=clusters,
            y=value,
            order=y_names,
            showfliers=show_fliers,
            fliersize=marker_size,
            orient=orient,
            whis=whis,
            flierprops={'marker': 'o', 'markersize': marker_size},
            boxprops={'linestyle': '-', 'linewidth': line_width},
            whiskerprops={'linestyle': '-', 'linewidth': line_width},
            medianprops={'linestyle': '-', 'linewidth': line_width},
            palette=palette if palette is not None else (colors if "color" in df_columns else None)
        )

        lines = ax1.lines

        for line in lines:
            line.set_linewidth(line_width)

        # set coordinate
        # noinspection DuplicatedCode
        ax.set_xticklabels(labels=y_names, rotation=rotation)
        ax.spines['top'].set_linewidth(line_width)
        ax.spines['bottom'].set_linewidth(line_width)
        ax.spines['left'].set_linewidth(line_width)
        ax.spines['right'].set_linewidth(line_width)

        if x_name is not None:
            plt.xlabel(x_name)

        plt.ylabel(y_name)

        plot_end(fig, output, show)


def box_trait(
    trait_df: DataFrame,
    trait_name: str = "All",
    trait_column_name: str = "id",
    value: str = "value",
    clusters: str = "clusters",
    x_name: str = None,
    y_name: str = "value",
    palette: Tuple = None,
    orient: str = None,
    width: float = 2,
    height: float = 2,
    line_width: float = 0.1,
    marker_size: float = 0.5,
    bottom: float = 0.3,
    rotation: float = 65,
    whis: float = 1.5,
    show_fliers: bool = True,
    is_sort: bool = True,
    order_names: list = None,
    title: str = None,
    output: path = None,
    show: bool = True
) -> None:

    data: DataFrame = trait_df.copy()

    def trait_plot(trait_: str, atac_cell_df_: DataFrame) -> None:
        """
        show plot
        :param trait_: trait name
        :param atac_cell_df_:
        :return: None
        """
        ul.log(__name__).info("Plotting box {}".format(trait_))
        # get gene score
        trait_score = atac_cell_df_[atac_cell_df_[trait_column_name] == trait_]
        # Sort gene scores from small to large
        box_base(
            df=trait_score,
            value=value,
            x_name=x_name,
            y_name=y_name,
            width=width,
            palette=palette,
            height=height,
            bottom=bottom,
            rotation=rotation,
            is_sort=is_sort,
            whis=whis,
            order_names=order_names,
            line_width=line_width,
            show_fliers=show_fliers,
            marker_size=marker_size,
            orient=orient,
            clusters=clusters,
            title=f"{title} {trait_}" if title is not None else title,
            output=os.path.join(output, f"cell_{trait_}_score_box.pdf") if output is not None else None,
            show=show
        )

    # noinspection DuplicatedCode
    trait_list = list(set(data[trait_column_name]))
    # judge trait
    if trait_name != "All" and trait_name not in trait_list:
        ul.log(__name__).error(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")
        raise ValueError(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")

    # plot
    if trait_name == "All":
        for trait in trait_list:
            trait_plot(trait, trait_df)
    else:
        trait_plot(trait_name, trait_df)



# -*- coding: UTF-8 -*-

import os
import sys

from . import file as fl
from . import model as ml
from . import plot as pl
from . import preprocessing as pp
from . import tool as tl
from . import util as ul

os.environ['OPENBLAS_NUM_THREADS'] = '1'
sys.setrecursionlimit(1000)

__version__ = f"{ul.project_name}: v{ul.project_version}"
__cache__ = ul.project_cache_path



# -*- coding: UTF-8 -*-

import os.path
from typing import Union, Tuple, Optional

import matplotlib
import numpy as np
from anndata import AnnData
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from pandas import DataFrame
import seaborn as sns

from .. import util as ul
from ..util import path, collection, type_50_colors, type_20_colors, chrtype, type_set_colors, plot_end

__name__: str = "plot_pie"

matplotlib.set_loglevel("error")


def scatter_base(
    df: DataFrame,
    x: str,
    y: str,
    hue: str = None,
    hue_order: list = None,
    title: str = None,
    bar_label: str = None,
    cmap: str = "Oranges",
    width: float = 2,
    height: float = 2,
    right: float = 0.9,
    text_fontsize: float = 7,
    legend_fontsize: float = 7,
    start_color_index: int = 0,
    color_step_size: int = 0,
    type_colors: collection = None,
    size: Union[float, collection] = 1.0,
    legend: dict = None,
    number: bool = False,
    is_text: bool = False,
    output: path = None,
    show: bool = True
) -> None:
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        fig, ax = plt.subplots(figsize=(width, height))
        fig.subplots_adjust(right=right)

        if title is not None:
            plt.title(title)

        # scatter
        if number:
            norm = plt.Normalize(df[hue].min(), df[hue].max())
            sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
            sm.set_array([])
            plt.colorbar(sm, label=bar_label)
            sns.scatterplot(
                data=df,
                x=x,
                y=y,
                palette=cmap,
                hue=hue,
                s=size,
                legend=False
            )
        else:
            __hue_order__ = list(np.sort(list(set(df[hue]))))

            if type_colors is None:
                type_colors = type_20_colors if len(__hue_order__) <= 20 else type_50_colors

            colors = {}

            if legend is not None:
                df.loc[:, "__hue__"] = df[hue].copy()

            i = 0
            for elem in __hue_order__:
                if legend is not None:
                    df.loc[df[df["__hue__"] == elem].index, "__hue__"] = legend[elem]
                    colors.update({legend[elem]: type_colors[start_color_index + i * color_step_size + __hue_order__.index(elem)]})
                else:
                    colors.update(
                        {
                            elem: type_colors[start_color_index + i * color_step_size + __hue_order__.index(elem)]
                        }
                    )
                i += 1

            if legend is not None:
                if hue_order is None:
                    hue_order = list(np.sort(list(set(df["__hue__"]))))
            else:
                if hue_order is None:
                    hue_order = __hue_order__

            sns.scatterplot(
                data=df,
                x=x,
                y=y,
                edgecolor=None,
                palette=colors,
                hue="__hue__" if legend is not None else hue,
                hue_order=hue_order,
                s=size
            )

            if is_text:

                df_anno = df[[hue, x, y]].groupby(hue, as_index=False).mean()

                for txt, i, j in zip(df_anno[hue], df_anno[x], df_anno[y]):
                    plt.annotate(
                        txt,
                        xy=(i, j),
                        xytext=(-10, 0),
                        textcoords="offset points",
                        bbox=dict(
                            boxstyle="round,pad=0.2",
                            fc="white",
                            ec="k",
                            lw=1,
                            alpha=0.8
                        ),
                        fontsize=text_fontsize
                    )

            ax.legend(
                loc="center left",
                bbox_to_anchor=(right, 0.5),
                bbox_transform=fig.transFigure,
                fontsize=legend_fontsize
            )

        # Remove scales and labels on the coordinate axis
        ax.set_xticks([])
        ax.set_yticks([])

        # Remove the bounding box of the coordinate axis
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['left'].set_visible(False)

        ax.set_xlabel(x)
        ax.set_ylabel(y)

        plot_end(fig, output, show)


def scatter_atac(
    adata: AnnData,
    columns: Tuple[str, str] = ("UMAP1", "UMAP2"),
    clusters: str = "clusters",
    hue_order: list = None,
    width: float = 2,
    height: float = 2,
    start_color_index: int = 0,
    color_step_size: int = 0,
    type_colors: collection = None,
    size: float = 1.0,
    text_fontsize: float = 7,
    legend_fontsize: float = 7,
    is_text: bool = False,
    output: path = None,
    show: bool = True
) -> None:
    # DataFrame
    df: DataFrame = adata.obs.copy()
    df[clusters] = df[clusters].astype(str)
    # scatter
    scatter_base(
        df,
        x=columns[0],
        y=columns[1],
        hue=clusters,
        width=width,
        height=height,
        size=size,
        hue_order=hue_order,
        start_color_index=start_color_index,
        color_step_size=color_step_size,
        type_colors=type_colors,
        is_text=is_text,
        text_fontsize=text_fontsize,
        legend_fontsize=legend_fontsize,
        output=output,
        show=show,
        right=0.75
    )


def scatter_trait(
    trait_adata: AnnData,
    title: str = None,
    bar_label: str = None,
    trait_name: str = "All",
    layers: Union[None, collection] = None,
    columns: Tuple[str, str] = ("UMAP1", "UMAP2"),
    cmap: str = "viridis",
    width: float = 2,
    height: float = 2,
    right: float = 0.9,
    number: bool = True,
    size: Union[float, collection] = 1.0,
    text_fontsize: float = 7,
    legend_fontsize: float = 7,
    start_color_index: int = 0,
    color_step_size: int = 0,
    type_colors: collection = None,
    is_text: bool = False,
    legend: dict = None,
    output: path = None,
    show: bool = True
) -> None:
    """
    Scatter plot of cell scores for traits/diseases
    :param legend_fontsize:
    :param text_fontsize:
    :param size:
    :param type_colors:
    :param right:
    :param height:
    :param width:
    :param legend:
    :param color_step_size:
    :param start_color_index:
    :param is_text:
    :param number:
    :param cmap:
    :param bar_label:
    :param title:
    :param columns:
    :param trait_adata: data
    :param trait_name: trait/disease name or All, 'All' show all traits/diseases
    :param layers: Matrix information used in drawing
    :param output: Image output path
    :param show: Whether to display pictures
    :return: None
    """

    data: AnnData = trait_adata.copy()

    # judge layers
    trait_adata_layers = list(data.layers)

    if layers is not None and len(layers) != 0:
        for layer in layers:
            if layer not in trait_adata_layers:
                ul.log(__name__).error("The `layers` parameter needs to include in `trait_adata.layers`")
                raise ValueError("The `layers` parameter needs to include in `trait_adata.layers`")

    def trait_plot(trait_: str, atac_cell_df_: DataFrame, layer_: str = None, new_data_: AnnData = None) -> None:
        """
        show plot
        :param trait_: trait name
        :param atac_cell_df_:
        :param layer_: layer
        :param new_data_:
        :return: None
        """
        ul.log(__name__).info(f"Plotting scatter {trait_}")
        # get gene score
        trait_score = new_data_[:, trait_].to_df()
        trait_score = trait_score.rename_axis("__barcode__")
        trait_score.reset_index(inplace=True)
        atac_cell_df_ = atac_cell_df_.rename_axis("__barcode__")
        atac_cell_df_.reset_index(inplace=True)
        # trait_score.rename_axis("index")
        df = atac_cell_df_.merge(trait_score, on="__barcode__", how="left")
        # Sort gene scores from small to large
        df.sort_values([trait_], inplace=True)
        scatter_base(
            df,
            x=columns[0],
            y=columns[1],
            hue=trait_,
            title=f"{title} {trait_}" if title is not None else title,
            bar_label=bar_label,
            legend=legend,
            cmap=cmap,
            width=width,
            height=height,
            right=right,
            number=number,
            size=size,
            type_colors=type_colors,
            text_fontsize=text_fontsize,
            legend_fontsize=legend_fontsize,
            start_color_index=start_color_index,
            color_step_size=color_step_size,
            is_text=is_text,
            output=os.path.join(
                output, f"cell_{trait_}_score_{layer_}.pdf" if layer_ is not None else f"cell_{trait_}_score.pdf"
            ) if output is not None else None,
            show=show
        )

    def handle_plot(layer_: str = None):
        # DataFrame
        atac_cell_df: DataFrame = data.obs.copy()
        atac_cell_df.rename_axis("index", inplace=True)
        trait_list: list = list(data.var_names)

        # judge trait
        if trait_name != "All" and trait_name not in trait_list:
            ul.log(__name__).error(f"The {trait_name} trait/disease is not in the trait/disease list (trait_adata.var_names)")
            raise ValueError(f"The {trait_name} trait/disease is not in the trait/disease list (trait_adata.var_names)")

        new_data: AnnData = AnnData(data.layers[layer], var=data.var, obs=data.obs) if layer_ is not None else data

        # plot
        if trait_name == "All":
            for trait in trait_list:
                trait_plot(trait, atac_cell_df, layer_, new_data)
        else:
            trait_plot(trait_name, atac_cell_df, layer_, new_data)

    if layers is None or len(layers) == 0:
        handle_plot()
    else:
        for layer in layers:
            ul.log(__name__).info(f"Start {layer}")
            handle_plot(layer)


def volcano_base(
    df: DataFrame,
    x: str = "Log2(Fold change)",
    y: str = "-Log10(P value)",
    hue: str = "type",
    size: int = 3,
    palette: Optional[list] = None,
    width: float = 2,
    height: float = 2,
    y_min: float = 0,
    axh_value: float = -np.log10(1e-3),
    axv_left_value: float = -1,
    axv_right_value: float = 1,
    title: str = None,
    x_name: Optional[str] = None,
    y_name: Optional[str] = None,
    output: path = None,
    show: bool = True
):
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:

        if palette is None:
            palette = ["#01c5c4", "#686d76", "#ff414d"]

        fig, ax = plt.subplots(figsize=(width, height))

        if title is not None:
            plt.title(title)

        sns.set_theme(style="ticks")
        sns.set_palette(sns.color_palette(palette))
        sns.scatterplot(data=df, x=x, y=y, hue=hue, s=size, ax=ax)
        ax.set_ylim(y_min, max(df[y]) * 1.1)

        if x_name is not None:
            ax.set_xlabel(x_name)

        if y_name is not None:
            ax.set_ylabel(y_name)

        plt.axhline(axh_value, color='grey', linestyle='--')
        plt.axvline(axv_left_value, color='grey', linestyle='--')
        plt.axvline(axv_right_value, color='grey', linestyle='--')

        # noinspection DuplicatedCode
        if output is not None:
            output_pdf = output if output.endswith(".pdf") else f"{output}.pdf"
            # plt.savefig(output_pdf, dpi=300)
            with PdfPages(output_pdf) as pdf:
                pdf.savefig(fig)

        if show:
            plt.grid(True)
            plt.show()

        plt.close()


def manhattan_causal_variant(
    df: DataFrame,
    y: str = "pp",
    chr_name: str = "chr",
    label: str = "rsId",
    size: int = 30,
    labels: Optional[list] = None,
    colors: Optional[list] = None,
    width: float = 8,
    height: float = 2,
    title: str = None,
    is_sort: bool = True,
    line_width: float = 0.5,
    y_round: int = 3,
    x_name: Optional[str] = "Chromosome",
    y_name: Optional[str] = "pp",
    y_limit: Tuple[float, float] = (0, 1),
    output: path = None,
    show: bool = True
):
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:

        df[chr_name] = df[chr_name].astype(chrtype)

        if is_sort:
            df = df.sort_values(chr_name)

        df['ind'] = range(len(df))
        df_grouped = df.groupby(chr_name)

        if colors is None:
            colors = type_20_colors.copy()
            colors.extend(type_set_colors)

        fig, ax = plt.subplots(figsize=(width, height))

        if title is not None:
            plt.title(title)

        x_labels = []
        x_labels_pos = []
        # Track the last index to draw lines between chromosomes
        last_ind = 0

        chr_unique = df[chr_name].unique()

        for num, (name, group) in enumerate(df_grouped):

            if name not in chr_unique:
                continue

            group.plot(kind='scatter', x='ind', y=y, color=colors[num], s=size, ax=ax)
            x_labels.append(name)
            x_labels_pos.append((group['ind'].iloc[-1] - (group['ind'].iloc[-1] - group['ind'].iloc[0]) / 2))

            # Draw a vertical line between chromosomes
            if num > 0:
                # Skip the first chromosome
                ax.axvline(x=last_ind + 0.5, color='gray', linestyle='--', linewidth=line_width)

            # Label specific mutations
            if labels is not None:
                for index, row in group.iterrows():
                    if row[label] in labels:
                        ax.text(row['ind'], row[y], row[label], ha='left', va='bottom')
                        ax.text(row['ind'], row[y], f"{y}={round(row[y], y_round)}", ha='left', va='top')

            last_ind = group['ind'].iloc[-1]

        # add grid
        ax.grid(axis="y", linestyle="--", linewidth=line_width, color="gray")
        ax.set_xticks(x_labels_pos)
        ax.set_xticklabels(x_labels)

        # Hide the borders above and to the right
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)

        ax.set_xlim([0, len(df)])
        ax.set_ylim(y_limit)

        if x_name is not None:
            ax.set_xlabel(x_name)

        if y_name is not None:
            ax.set_ylabel(y_name)

        plot_end(fig, output, show)



# -*- coding: UTF-8 -*-

from typing import Optional, Union, Tuple

import pandas as pd
from anndata import AnnData
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from PyComplexHeatmap import HeatmapAnnotation, anno_simple, ClusterMapPlotter, anno_label, anno_barplot
from matplotlib.colors import ListedColormap
from pandas import DataFrame
import seaborn as sns

from .. import util as ul
from ..util import path, type_20_colors, type_50_colors, plot_end

__name__: str = "plot_heat_map"


def heatmap_annotation(
    adata: AnnData,
    layer: Optional[str] = None,
    width: float = 4,
    height: float = 4,
    title: Optional[str] = None,
    label: str = "value",
    row_name: Optional[str] = None,
    col_name: Optional[str] = None,
    row_names: Optional[str] = None,
    col_names: Optional[str] = None,
    row_anno_label: bool = False,
    col_anno_label: bool = False,
    row_anno_text: bool = False,
    col_anno_text: bool = False,
    row_legend: bool = False,
    col_legend: bool = False,
    row_show_names: bool = False,
    col_show_names: bool = False,
    row_cluster: bool = False,
    col_cluster: bool = False,
    cluster_method: str = "average",
    cluster_metric: str = "correlation",
    row_names_side: str = "left",
    col_names_side: str = "bottom",
    top: float = 0.05,
    right: float = 0.15,
    bottom: float = 0.05,
    left: float = 0.05,
    label_size: float = 9,
    fontsize: float = 9,
    is_position: bool = False,
    level_bar_height: float = None,
    anno_specific_labels: list = None,
    x_label_rotation: float = 245,
    y_label_rotation: float = 0,
    row_color_start_index: int = 0,
    col_color_start_index: int = 10,
    row_split: Union[int, pd.Series] = None,
    col_split: Union[int, pd.Series] = None,
    row_split_order: Union[list, str] = None,
    col_split_order: Union[list, str] = None,
    row_split_gap: float = 0.5,
    col_split_gap: float = 0.2,
    frac: float = 0.2,
    relpos: Tuple = (0, 1),
    anno_label_height: Optional[float] = None,
    selected_anno_label_height: float = 2.5,
    category_height: Optional[float] = 2.5,
    x_name: Optional[str] = None,
    y_name: Optional[str] = None,
    row_score_name: str = "association_score",
    cmap: str = "Oranges",
    is_sort: bool = True,
    show: bool = True,
    output: path = None,
    **kwargs
) -> None:
    """
    Generate a heatmap with row and column annotations.
    :param row_score_name:
    :param category_height:
    :param selected_anno_label_height:
    :param anno_label_height:
    :param frac:
    :param relpos:
    :param level_bar_height:
    :param anno_specific_labels:
    :param adata: input data;
    :param layer: Layer for processing data;
    :param width: The width of the file or image;
    :param height: The height of the file or image;
    :param title: Title of the image;
    :param label: Explanation (title) of the heatmap display value icon in the figure;
    :param row_name: Information on row annotations in the figure;
    :param col_name: Information on column annotations in the figure;
    :param row_names: The information of the row in the picture;
    :param col_names: The information of the column in the figure;
    :param row_anno_label: Whether to display the label of row comments;
    :param col_anno_label: Whether to display the label of column comments;
    :param row_anno_text: Whether to display row comment information in row comments;
    :param col_anno_text: Whether to display column comment information in column comments;
    :param row_legend: Whether to display the category description of the row in the row comments;
    :param col_legend: Whether to display the category description of the column in the column comments;
    :param row_show_names: If set to `true`, display the name of each element in the row;
    :param col_show_names: If set to `true`, display the name of each element in the column;
    :param row_cluster: Set to `true`, perform row clustering;
    :param col_cluster: Set to `true`, perform column clustering;
    :param cluster_method: If `row_cluster` or `col_cluster` is true, the clustering method will be used;
    :param cluster_metric: If `row_cluster` or `col_cluster` is true, the clustering method will be used;
    :param row_names_side: Direction of row names display. Effective when `row_names_side` is `true`;
    :param col_names_side: Direction of column names display. Effective when `col_show_names` is `true`;
    :param top: The gap at the top in the picture. Effective when `is_position` is `true`;
    :param right: The gap at the right in the picture. Effective when `is_position` is `true`;
    :param bottom: The gap at the bottom in the picture. Effective when `is_position` is `true`;
    :param left: The gap at the left in the picture. Effective when `is_position` is `true`;
    :param label_size: The size of the font for row or column names. Effective when `row_names_side` or `col_show_names` is `true`;
    :param fontsize: The size of the row or column title. Effective when `x_name` or `y_name` not is `None`;
    :param is_position: Set to true, customize the distance of surrounding gaps in the image;
    :param x_label_rotation: The degree of rotation for row names. Effective when `row_names_side` is `true`;
    :param y_label_rotation: The degree of rotation for column names. Effective when `col_show_names` is `true`;
    :param row_color_start_index: Row annotation specifies the starting index of different colors;
    :param col_color_start_index: Column annotation specifies the starting index of different colors;
    :param row_split_order: list or str
        a list to specify the order of row_split, could also be 'cluster_between_groups', if cluster_between_groups was specified,
        hierarchical clustering will be performed on the mean values for each groups and pass the clustered order to row_split_order.
        For example,see https://dingwb.github.io/PyComplexHeatmap/build/html/notebooks/advanced_usage.html#Cluster-between-groups
    :param col_split_order: list or str
        a list to specify the order of col_split, could also be 'cluster_between_groups', if cluster_between_groups was specified,
        hierarchical clustering will be performed on the mean values for each groups and pass the clustered order to row_split_order.
    :param row_split_gap: default are 0.5 and 0.2 mm for row and col.
    :param col_split_gap: default are 0.5 and 0.2 mm for row and col.
    :param row_split: int or pd.Series or pd.DataFrame
        number of cluster for hierarchical clustering or pd.Series or pd.DataFrame, used to split rows or rows into subplots.
    :param col_split: int or pd.Series or pd.DataFrame
        int or pd.Series or pd.DataFrame, used to split rows or columns into subplots.
    :param x_name: Title of row name;
    :param y_name: Title of column name;
    :param cmap: Display color themes for heat maps;
    :param is_sort: If set to true, when displaying the heatmap, the row and column names are sorted and displayed;
    :param show: If true, display the image;
    :param output: Output file for image saving;
    :return: Display of image or saved file.
    """

    if output is None and not show:
        ul.log(__name__).info(f"At least one of the `output` and `show` parameters is required.")
    else:
        ul.log(__name__).info("Start plotting the heatmap")
        # noinspection DuplicatedCode
        fig, ax = plt.subplots(figsize=(width, height))
        data = adata.copy()

        # judge layers
        if layer is not None:

            if layer not in list(data.layers):
                ul.log(__name__).error("The `layer` parameter needs to include in `adata.layers`")
                raise ValueError(f"The `{layer}` parameter needs to include in `adata.layers`")

            data.X = data.layers[layer]

        if is_sort:
            data = data[data.obs.sort_values(row_name).index, data.var.sort_values(col_name).index]

        # DataFrame
        df: DataFrame = data.to_df()

        row_anno: DataFrame = data.obs.copy()

        col_anno: DataFrame = data.var.copy()

        if row_names is not None:
            df.index = data.obs[row_names].astype(str)
            row_anno.index = data.obs[row_names].astype(str)

        if col_names is not None:
            df.columns = data.var[col_names].astype(str)
            col_anno.index = data.var[col_names].astype(str)

        if is_position:
            plt.subplots_adjust(left=left, right=1 - right, top=1 - top, bottom=bottom)

        if row_name is not None:
            row_colors = type_20_colors[row_color_start_index:] if len(list(set(row_anno[row_name]))) + row_color_start_index <= 20 else type_50_colors[row_color_start_index:]
        else:
            row_colors = "cmap50"

        if col_name is not None:
            col_colors = type_20_colors[col_color_start_index:] if len(list(set(col_anno[col_name]))) + col_color_start_index <= 20 else type_50_colors[col_color_start_index:]
        else:
            col_colors = "cmap50"

        if title is not None:
            plt.title(title)

        df_rows = None
        if anno_specific_labels is not None:
            df_rows = df.apply(lambda x: x.name if x.name in anno_specific_labels else None, axis=1)
            df_rows.name = "Selected"

        # noinspection PyTypeChecker
        row_ha = HeatmapAnnotation(
            label=anno_label(
                row_anno[row_name], cmap=ListedColormap(row_colors), merge=True, height=anno_label_height
            ) if row_anno_label else None,
            RowCategory=anno_simple(
                row_anno[row_name],
                cmap=ListedColormap(row_colors),
                height=category_height,
                legend=row_legend,
                add_text=row_anno_text,
                text_kws=dict(color="black", rotation=0, fontsize=label_size),
            ) if row_name is not None else None,
            axis=0,
            verbose=0,
            legend_gap=5,
            hgap=0.5,
            label_kws=dict(color="black", rotation=90, horizontalalignment="left")
        )

        # noinspection PyTypeChecker
        row_ha_right = HeatmapAnnotation(
            AssociationScore=anno_barplot(row_anno[[row_score_name]], legend=True, height=level_bar_height, **dict(edgecolor='none')) if row_score_name in row_anno.columns else None,
            selected=anno_label(df_rows, relpos=relpos, frac=frac, height=selected_anno_label_height) if anno_specific_labels is not None else None,
            axis=0,
            verbose=0,
            legend_gap=5,
            hgap=0.5,
            label_kws=dict(color="black", rotation=90, horizontalalignment="left")
        )

        col_ha_args = {"rotation": 90}
        # noinspection PyTypeChecker
        col_ha = HeatmapAnnotation(
            label=anno_label(
                col_anno[col_name], cmap=ListedColormap(col_colors), merge=True, height=anno_label_height, **col_ha_args
            ) if col_anno_label else None,
            ColCategory=anno_simple(
                col_anno[col_name],
                cmap=ListedColormap(col_colors),
                height=category_height,
                add_text=col_anno_text,
                legend=col_legend,
                text_kws={'fontsize': label_size}
            ) if col_name is not None else None,
            axis=1,
            verbose=0,
            legend_gap=5,
            hgap=0.5,
            label_side='left',
            label_kws=dict(color="black", rotation=0, horizontalalignment="right")
        )

        """
        It is worth noting here that, `row_cluster_metric="correlation"`, When the default parameter 
        `row_cluster_metric` in method `ClusterMapPlotter` is passed into method `distance.pdist`, 
        that is `metric='correlation'`, and this method derives from this 
        `https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.spatial.distance.pdist.html`,
        it can be inferred that there is a division formula in one step, which may result in the possibility of `NA`.
        
        For example, in this scATAC-seq data, if there are two or more traits without any intersection, 
        the denominator will appear as zero.
        
        Therefore, use the `"median"` value for parameter `row_cluster_method`
        Therefore, use the `"euclidean"` value for parameter `row_cluster_metric`
        """

        ClusterMapPlotter(
            data=df,
            top_annotation=col_ha if col_name is not None else None,
            left_annotation=row_ha if row_name is not None else None,
            right_annotation=row_ha_right if anno_specific_labels is not None or row_score_name in row_anno.columns else None,
            label=label,
            row_cluster_method=cluster_method,
            row_cluster_metric=cluster_metric,
            col_cluster_method=cluster_method,
            col_cluster_metric=cluster_metric,
            show_rownames=row_show_names,
            show_colnames=col_show_names,
            row_names_side=row_names_side,
            col_names_side=col_names_side,
            col_split=col_split,
            row_split=row_split,
            row_split_order=row_split_order,
            col_split_order=col_split_order,
            col_split_gap=col_split_gap,
            row_split_gap=row_split_gap,
            xticklabels_kws=dict(labelrotation=x_label_rotation, labelcolor='black', labelsize=label_size),
            yticklabels_kws=dict(labelrotation=y_label_rotation, labelcolor='black', labelsize=label_size),
            cmap=cmap,
            tree_kws={'row_cmap': 'Dark2'},
            xlabel=x_name,
            ylabel=y_name,
            xlabel_kws=dict(color='black', fontsize=fontsize),
            ylabel_kws=dict(color='black', fontsize=fontsize),
            col_cluster=col_cluster,
            row_cluster=row_cluster,
            col_dendrogram=col_cluster,
            row_dendrogram=row_cluster,
            **kwargs
        )

        plot_end(fig, output, show)


def heatmap(
    adata: AnnData,
    layer: str = None,
    width: float = 4,
    height: float = 4,
    annot: bool = False,
    square: bool = True,
    is_cluster: bool = False,
    cmap: str = "Oranges",
    line_widths: float = 1,
    fmt: str = ".2f",
    rotation: float = 65,
    x_name: str = None,
    y_name: str = None,
    output: path = None,
    show: bool = True
) -> None:
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        # noinspection DuplicatedCode
        fig, ax = plt.subplots(figsize=(width, height))
        data = adata.copy()

        # judge layers
        if layer is not None:

            if layer not in list(data.layers):
                ul.log(__name__).error("The `layer` parameter needs to include in `adata.layers`")
                raise ValueError("The `layer` parameter needs to include in `adata.layers`")

            data.X = data.layers[layer]

        # DataFrame
        ul.log(__name__).info(f"to DataFrame")
        df: DataFrame = data.to_df()
        # seaborn
        heat_map: Axes = sns.clustermap(data=df, square=square, annot=annot, cmap=cmap, fmt=fmt) \
            if is_cluster else \
            sns.heatmap(data=df, square=square, annot=annot, cmap=cmap, linewidths=line_widths, fmt=fmt)

        if not is_cluster:
            plt.subplots_adjust(left=0, bottom=0)
            plt.setp(
                heat_map.get_xticklabels(), rotation=rotation, ha="right", rotation_mode="anchor"
            )
        else:
            # noinspection PyUnresolvedReferences
            plt.setp(heat_map.ax_heatmap.get_xticklabels(), rotation=rotation)

        if x_name is not None:
            plt.xlabel(x_name)

        if y_name is not None:
            plt.ylabel(y_name)

        plot_end(fig, output, show)



# -*- coding: UTF-8 -*-

import os
from typing import Tuple, Union, Literal

from matplotlib import pyplot as plt
from pandas import DataFrame
import seaborn as sns

from .. import util as ul
from ..util import path, plot_end

__name__: str = "plot_violin"

_Kind = Literal["strip", "swarm", "box", "violin", "boxen", "point", "bar", "count"]


def violin_base(
    df: DataFrame,
    value: str = "value",
    x_name: str = None,
    y_name: str = "value",
    kind: _Kind = "violin",
    clusters: str = "clusters",
    palette: Union[Tuple, list] = None,
    hue: str = None,
    width: float = 2,
    height: float = 2,
    bottom: float = 0.3,
    rotation: float = 65,
    line_width: float = 0.5,
    title: str = None,
    split: bool = True,
    is_sort: bool = True,
    order_names: list = None,
    output: path = None,
    show: bool = True
) -> None:
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
        return

    # judge
    # noinspection DuplicatedCode
    df_columns = list(df.columns)

    if value not in df_columns:
        ul.log(__name__).error(f"The `value` ({value}) parameter must be in the `df` parameter data column name ({df_columns})")
        raise ValueError(f"The `value` ({value}) parameter must be in the `df` parameter data column name ({df_columns})")

    if hue is not None and hue not in df_columns:
        ul.log(__name__).error(f"The `hue` ({hue}) parameter must be in the `df` parameter data column name ({df_columns})")
        raise ValueError(f"The `hue` ({hue}) parameter must be in the `df` parameter data column name ({df_columns})")

    # noinspection DuplicatedCode
    fig, ax = plt.subplots(figsize=(width, height))
    fig.subplots_adjust(bottom=bottom)

    if title is not None:
        plt.title(title)

    group_columns = [clusters]

    new_df: DataFrame = df.groupby(group_columns, as_index=False)[value].median()

    if "color" in df_columns:
        new_df_color: DataFrame = df.groupby(group_columns, as_index=False)["color"].first()
        new_df = new_df.merge(new_df_color, how="left", on=clusters)

    colors: list = []

    # sort
    if is_sort:
        new_df.sort_values([value], ascending=False, inplace=True)
        y_names: Union[list, None] = list(new_df[clusters])
        if "color" in df_columns:
            colors = list(new_df["color"])
    else:
        new_df.index = new_df[clusters]
        if order_names is not None:
            y_names: list = order_names
            if "color" in df_columns:
                for i in order_names:
                    for j, c in zip(new_df[clusters], new_df["color"]):
                        if i == j:
                            colors.append(c)
                            break
        else:
            y_names = list(new_df[clusters])
            if "color" in df_columns:
                colors = list(new_df["color"])

    # scatter
    g = sns.catplot(
        data=df,
        x=clusters,
        y=value,
        kind=kind,
        hue=hue,
        order=y_names,
        split=split,
        linewidth=line_width,
        palette=palette if palette is not None else (colors if "color" in df_columns else None)
    )

    # set coordinate
    for ax in g.axes.flat:
        ax.spines['top'].set_linewidth(line_width)
        ax.spines['right'].set_linewidth(line_width)
        ax.spines['bottom'].set_linewidth(line_width)
        ax.spines['left'].set_linewidth(line_width)
        # Set the rotation angle of the x-axis labels
        ax.tick_params(axis='x', rotation=rotation)

    if x_name is not None:
        plt.xlabel(x_name)

    plt.ylabel(y_name)

    plot_end(fig, output, show)


def violin_trait(
    trait_df: DataFrame,
    trait_name: Union[str, list] = "All",
    trait_column_name: str = "id",
    value: str = "value",
    clusters: str = "clusters",
    kind: _Kind = "violin",
    x_name: str = None,
    y_name: str = "value",
    palette: Tuple = None,
    width: float = 2,
    height: float = 2,
    rotation: float = 65,
    line_width: float = 0.1,
    bottom: float = 0.3,
    split: bool = True,
    is_sort: bool = True,
    order_names: list = None,
    title: str = None,
    output: path = None,
    show: bool = True
) -> None:
    """
    Violin plot of cell scores for traits/diseases
    :param palette:
    :param is_sort:
    :param order_names:
    :param bottom:
    :param rotation:
    :param line_width:
    :param height:
    :param kind:
    :param x_name:
    :param clusters:
    :param width:
    :param title:
    :param y_name:
    :param split:
    :param value:
    :param trait_column_name:
    :param trait_df: data
    :param trait_name: trait/disease name or All, 'All' show all traits/diseases
    :param output: Image output path
    :param show: Whether to display pictures
    :return: None
    """

    data: DataFrame = trait_df.copy()

    def trait_plot(trait_: Union[str, list], atac_cell_df_: DataFrame) -> None:
        """
        show plot
        :param trait_: trait name
        :param atac_cell_df_:
        :return: None
        """
        ul.log(__name__).info("Plotting box {}".format(", ".join(trait_)))
        # get gene score
        _filename_: str = trait_ if isinstance(trait_, str) else "_".join(trait_)
        trait_score = atac_cell_df_[atac_cell_df_[trait_column_name] == trait_ if isinstance(trait_, str) else atac_cell_df_[trait_column_name].isin(trait_)]
        # Sort gene scores from small to large
        violin_base(
            df=trait_score,
            value=value,
            x_name=x_name,
            y_name=y_name,
            width=width,
            palette=palette,
            height=height,
            bottom=bottom,
            split=split,
            is_sort=is_sort,
            rotation=rotation,
            order_names=order_names,
            kind=kind,
            hue=trait_column_name,
            line_width=line_width,
            clusters=clusters,
            title=f"{title} {_filename_}" if title is not None else title,
            output=os.path.join(output, f"cell_{_filename_}_score_cat_{kind}.pdf") if output is not None else None,
            show=show
        )

    # noinspection DuplicatedCode
    trait_list = list(set(data[trait_column_name]))
    # judge trait
    if trait_name != "All":
        if isinstance(trait_name, str) and trait_name not in trait_list:
            ul.log(__name__).error(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")
            raise ValueError(f"The {trait_name} trait/disease is not in the trait/disease list {trait_list}.")
        else:
            for tn in trait_name:
                if tn not in trait_list:
                    ul.log(__name__).error(f"The {tn} trait/disease is not in the trait/disease list {trait_list}.")
                    raise ValueError(f"The {tn} trait/disease is not in the trait/disease list {trait_list}.")

    # plot
    if trait_name == "All":
        for trait in trait_list:
            trait_plot(trait, trait_df)
    else:
        trait_plot(trait_name, trait_df)



# -*- coding: UTF-8 -*-

import math
import random
from typing import Optional, Literal, Union

import matplotlib
import networkx as nx
import numpy as np
from anndata import AnnData
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib import pyplot as plt, gridspec

from .. import util as ul
from ..util import (
    matrix_data,
    collection,
    path,
    list_index,
    to_dense,
    type_20_colors,
    type_50_colors,
    check_adata_get,
    plot_end
)

__name__: str = "plot_graph"

_LayoutType = Optional[Literal['spring', 'kamada_kawai', 'circular', 'shell', 'circular_type1', 'circular_type2', 'square_type1', 'square_type2']]


def graph(
    data: matrix_data,
    labels: collection = None,
    node_size: int = 50,
    width: float = 2,
    height: float = 2,
    is_font: bool = False,
    output: path = None,
    show: bool = True
) -> None:
    """
    plot graph from adjacency matrix
    :param data: input data
    :param labels: Label of data node
    :param node_size: Size of data node
    :param width:
    :param height:
    :param is_font: Is there a word on the data node
    :param output: Image output path
    :param show: Whether to display pictures
    :return: None
    """
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        plt.figure(figsize=(width, height), dpi=150)

        # Determine whether it is a square array
        if data.shape[0] != data.shape[1]:
            ul.log(__name__).error("The input data must be a square matrix.")
            raise ValueError("The input data must be a square matrix.")

        # set labels
        labels_dict = {}

        if labels is not None:

            if data.shape[0] != np.asarray(labels).size:
                ul.log(__name__).error(
                    f"The number of input data nodes {data.shape[0]} and the number of "
                    f"labels {np.asarray(labels).size} must be consistent"
                )
                raise ValueError(
                    f"The number of input data nodes {data.shape[0]} and the number of "
                    f"labels {np.asarray(labels).size} must be consistent"
                )

            labels_dict: dict = dict(zip(range(len(labels)), labels))

        rows, cols = np.where(data == 1)
        edges = zip(rows.tolist(), cols.tolist())
        gr = nx.Graph(name="Cell-cell graph")
        gr.add_edges_from(edges)
        pos = nx.spring_layout(gr, k=0.15, iterations=35, seed=2023)

        options: dict = {
            "node_color": "black",
            "node_size": node_size,
            "linewidths": 0,
            "width": 0.1
        }

        if is_font:
            if labels is not None:
                nx.draw(gr, pos=pos, labels=labels_dict, **options)
            else:
                nx.draw(gr, pos=pos, **options)
        else:
            nx.draw(gr, pos=pos, labels={}, **options)

        if show:
            plt.show()

        if output is not None:
            plt.savefig(output if output.endswith(".pdf") else f"{output}.pdf")

        plt.close()


def communities_graph(
    adata: AnnData,
    labels: collection,
    layer: str = None,
    clusters: str = "clusters",
    width: float = 2,
    height: float = 2,
    title: str = None,
    node_size: float = 2.0,
    line_widths: float = 0.001,
    start_color_index: int = 0,
    color_step_size: int = 0,
    output: path = None,
    show: bool = True
):
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        ul.log(__name__).info("Start cell-cell network diagram")

        new_data = check_adata_get(adata=adata, layer=layer)

        # adjust matrix
        adj_matrix = to_dense(new_data.X)
        communities, node_labels = list_index(labels)

        df = new_data.obs.copy()

        __hue_order__ = list(np.sort(list(set(df[clusters]))))

        type_colors = type_20_colors if len(__hue_order__) <= 20 else type_50_colors

        ul.log(__name__).info("Get position")
        color_index = 0
        g = nx.from_numpy_array(adj_matrix)
        partition: list = [0 for _ in range(g.number_of_nodes())]

        for c_i, nodes in enumerate(communities):
            for i in nodes:
                partition[i] = type_colors[start_color_index + color_index * color_step_size + c_i]
            color_index += 1

        pos = nx.spring_layout(g)

        pos1 = [p[0] for p in pos.values()]
        pos2 = [p[1] for p in pos.values()]
        new_data.obs["pos1"] = pos1
        new_data.obs["pos2"] = pos2

        fig = plt.figure(figsize=(width, height))

        if title is not None:
            plt.title(title)

        plt.axis("off")

        nx.draw_networkx_nodes(
            g,
            pos=pos,
            node_size=node_size,
            node_color=partition,
            linewidths=line_widths
        )
        # nodes.set_edgecolor("b")
        nx.draw_networkx_edges(
            g,
            pos=pos,
            node_size=node_size,
            edge_color=(0, 0, 0, 0.3),
            width=line_widths
        )

        # noinspection DuplicatedCode
        if output is not None:
            output_pdf = output if output.endswith(".pdf") else f"{output}.pdf"
            # plt.savefig(output_pdf, dpi=300)
            with PdfPages(output_pdf) as pdf:
                pdf.savefig(fig)

        if show:
            plt.grid(True)
            plt.show()

        plt.close()


def network_two_types(
    data_pairs: list,
    type1_scores: dict,
    type2_scores: dict,
    type1_node_size: Optional[Union[dict, list, float]] = 50,
    type2_node_size: Optional[Union[dict, list, float]] = 50,
    label_nodes: Optional[list] = None,
    width: float = 4,
    height: float = 3,
    k: Optional[float] = None,
    iterations: int = 50,
    scale: float = 1,
    radius: float = 0.35,
    type1_node_shape: str = 'o',
    type2_node_shape: str = 's',
    type1_bar_label: str = 'Score',
    type2_bar_label: str = 'Score',
    type1_cmap_str: str = "winter",
    type2_cmap_str: str = "YlOrRd",
    node_alpha: float = 0.8,
    edge_alpha: float = 0.8,
    is_fluctuate: bool = True,
    layout_type: str = 'spring',
    output: path = None,
    show: bool = True
):

    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        # 创建基因和变异的节点列表
        type1_nodes = []
        type2_nodes = []

        type1_keys = type1_scores.keys()
        type2_keys = type2_scores.keys()

        for type1_node, type2_node in data_pairs:

            if type1_node not in type1_keys:
                raise Exception(f"Node `{type1_node}` not in `type1_scores`.")

            if type2_node not in type2_keys:
                raise Exception(f"Node `{type2_node}` not in `type2_scores`.")

            if type1_node not in type1_nodes:
                type1_nodes.append(type1_node)

            if type2_node not in type2_nodes:
                type2_nodes.append(type2_node)

        # 创建图形和网格
        fig = plt.figure(figsize=(width, height))
        gs = gridspec.GridSpec(
            2, 2,
            width_ratios=[7, 1],
            height_ratios=[4, 4],
            hspace=0.1,
            wspace=0.1
        )

        # 主轴用于绘制网络图
        ax_network = fig.add_subplot(gs[0:2, 0:1])

        # 创建有向图
        G = nx.Graph()
        G.add_edges_from(data_pairs)

        # 绘制图形
        if layout_type == 'spring':
            pos = nx.spring_layout(G, k=k, iterations=iterations)
        elif layout_type == 'kamada_kawai':
            pos = nx.kamada_kawai_layout(G, scale=scale)
        elif layout_type == 'circular':
            pos = nx.circular_layout(G, scale=scale)
        elif layout_type == 'shell':
            shells = [type1_nodes, type2_nodes]
            pos = nx.shell_layout(G, shells, scale=scale)
        else:
            if layout_type == 'circular_type1':
                is_type1 = True
                type1_positions = nx.circular_layout(G.subgraph(type1_nodes), scale=scale)
                pos = {**type1_positions}
            elif layout_type == 'circular_type2':
                is_type1 = False
                type2_positions = nx.circular_layout(G.subgraph(type2_nodes), scale=scale)
                pos = {**type2_positions}
            else:
                if layout_type == 'square_type1':
                    is_type1 = True
                elif layout_type == 'square_type2':
                    is_type1 = False
                else:
                    raise ValueError("The `layout_type` parameter must be one of the following string values {'spring','kamada_kawai','circular','shell','circular_type1','circular_type2','square_type1','square_type2'}")

                # 调整为正方形坐标
                pos = {}
                side_length = int(np.sqrt(len(type1_nodes if is_type1 else type2_nodes)))
                for i, key in enumerate(type1_nodes if is_type1 else type2_nodes):
                    pos[key] = (i % side_length, i // side_length)

            for i, _node_ in enumerate(type1_nodes if is_type1 else type2_nodes):
                type_other = []
                for k, v in data_pairs:
                    if is_type1:
                        if k == _node_:
                            type_other.append(v)
                    else:
                        if v == _node_:
                            type_other.append(k)

                angle_step = 360 / len(type_other)
                for j, _type_node_ in enumerate(type_other):
                    angle = math.radians(j * angle_step) + (random.random() / 2 if is_fluctuate else 0)
                    x = pos[_node_][0] + np.power(-1, int(random.random() * 10) if is_fluctuate else 0) * (radius + (random.random() / 10 if is_fluctuate else 0)) * math.cos(angle)
                    y = pos[_node_][1] + np.power(-1, int(random.random() * 10) if is_fluctuate else 0) * (radius + (random.random() / 10 if is_fluctuate else 0)) * math.sin(angle)
                    pos[_type_node_] = (x, y)

        type1_cmap = matplotlib.colormaps[type1_cmap_str]
        type2_cmap = matplotlib.colormaps[type2_cmap_str]

        # 确保颜色映射的归一化范围与基因分数的值范围相匹配
        type1_norm = plt.Normalize(vmin=min(type1_scores.values()), vmax=max(type1_scores.values()))
        type2_norm = plt.Normalize(vmin=min(type2_scores.values()), vmax=max(type2_scores.values()))

        type1_node_colors = [type1_cmap(type1_norm(type1_scores[node])) for node in type1_nodes]
        nx.draw_networkx_nodes(
            G,
            pos,
            nodelist=type1_nodes,
            node_size=[type1_node_size[node] for node in type1_nodes] if isinstance(type1_node_size, dict) else type1_node_size,
            node_color=type1_node_colors,
            edgecolors='#333333',
            node_shape=type1_node_shape,
            alpha=node_alpha,
            ax=ax_network
        )

        type2_node_colors = [type2_cmap(type2_norm(type2_scores[node])) for node in type2_nodes]
        nx.draw_networkx_nodes(
            G,
            pos,
            nodelist=type2_nodes,
            node_size=[type2_node_size[node] for node in type2_nodes] if isinstance(type2_node_size, dict) else type2_node_size,
            node_color=type2_node_colors,
            edgecolors='#333333',
            node_shape=type2_node_shape,
            alpha=node_alpha,
            ax=ax_network
        )

        if label_nodes is not None:
            labels = {node: node for node in label_nodes}

            # 绘制节点标签
            nx.draw_networkx_labels(
                G,
                pos,
                labels,
                font_size=10,
                ax=ax_network,
                verticalalignment='center',
                horizontalalignment='center'
            )

        # 边的颜色和宽度
        edge_colors = ['gray' for _ in data_pairs]
        edge_widths = [1.0 for _ in data_pairs]

        # 绘制边
        nx.draw_networkx_edges(
            G,
            pos,
            edge_color=edge_colors,
            width=edge_widths,
            alpha=edge_alpha,
            arrows=False,
            ax=ax_network,
            style='solid'
        )

        # 设置背景颜色
        ax_network.set_facecolor('white')
        ax_network.spines['top'].set_visible(False)
        ax_network.spines['right'].set_visible(False)

        # 颜色条轴
        ax_cbar1 = fig.add_subplot(gs[0:1, 1:2])
        ax_cbar1.set_axis_off()
        # 创建颜色条
        sm1 = plt.cm.ScalarMappable(cmap=type1_cmap, norm=type1_norm)
        cbar1 = plt.colorbar(sm1, ax=ax_cbar1, location='right', pad=0.2)
        cbar1.set_label(type1_bar_label)

        ax_cbar2 = fig.add_subplot(gs[1:2, 1:2])
        ax_cbar2.set_axis_off()
        sm2 = plt.cm.ScalarMappable(cmap=type2_cmap, norm=type2_norm)
        cbar2 = plt.colorbar(sm2, ax=ax_cbar2, location='right', pad=0.2)
        cbar2.set_label(type2_bar_label)

        # 显示图形
        plt.axis('off')

        plot_end(fig, output, show)



# -*- coding: UTF-8 -*-

from ._algorithm_ import (
    sigmoid,
    tf_idf,
    z_score_normalize,
    z_score_marginal,
    marginal_normalize,
    min_max_norm,
    symmetric_scale,
    mean_symmetric_scale,
    coefficient_of_variation,
    is_asc_sort,
    lsi,
    pca,
    jaccard_similarity,
    spectral_eigenmaps,
    semi_mutual_knn_weight,
    k_means,
    spectral_clustering,
    tsne,
    umap,
    kl_divergence,
    safe_kl_divergence,
    calinski_harabasz,
    silhouette,
    davies_bouldin,
    ari,
    ami,
    binary_indicator,
    z_score_to_p_value,
    euclidean_distances,
    overlap,
    overlap_sum,
    calculate_fragment_weighted_accessibility,
    calculate_init_score_weight,
    obtain_cell_cell_network,
    perturb_data,
    add_noise
)

from ._random_walk_ import (
    trs_scale_norm,
    TraitDataParallel,
    random_walk,
    RandomWalk,
)

__all__ = [
    "sigmoid",
    "tf_idf",
    "z_score_normalize",
    "z_score_marginal",
    "marginal_normalize",
    "min_max_norm",
    "symmetric_scale",
    "mean_symmetric_scale",
    "coefficient_of_variation",
    "is_asc_sort",
    "lsi",
    "pca",
    "jaccard_similarity",
    "spectral_eigenmaps",
    "semi_mutual_knn_weight",
    "k_means",
    "spectral_clustering",
    "tsne",
    "umap",
    "kl_divergence",
    "safe_kl_divergence",
    "calinski_harabasz",
    "silhouette",
    "davies_bouldin",
    "ari",
    "ami",
    "binary_indicator",
    "z_score_to_p_value",
    "trs_scale_norm",
    "TraitDataParallel",
    "random_walk",
    "RandomWalk",
    "euclidean_distances",
    "overlap",
    "overlap_sum",
    "calculate_fragment_weighted_accessibility",
    "calculate_init_score_weight",
    "obtain_cell_cell_network",
    "perturb_data",
    "add_noise"
]



# -*- coding: UTF-8 -*-

from typing import Literal

import numpy as np
from anndata import AnnData
from matplotlib import pyplot as plt
import seaborn as sns
from tqdm import tqdm

from .. import util as ul
from ..util import path, down_sampling_data, check_adata_get, plot_end

__name__: str = "plot_kde"


def kde(
    adata: AnnData,
    layer: str = None,
    title: str = None,
    width: float = 4,
    height: float = 2,
    axis: Literal[-1, 0, 1] = -1,
    sample_number: int = 1000000,
    is_legend: bool = True,
    output: path = None,
    show: bool = True
) -> None:
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        ul.log(__name__).info("Start plotting the Kernel density estimation chart")
        fig, ax = plt.subplots(figsize=(width, height))
        fig.subplots_adjust(bottom=0.3)

        data = check_adata_get(adata, layer=layer, is_dense=True, is_matrix=False)

        if title is not None:
            plt.title(title)

        sns.set_theme(style="whitegrid")

        # Random sampling
        if axis == -1:
            matrix = down_sampling_data(data.X, sample_number)
            sns.kdeplot(matrix, shade=True, fill=True)
        elif axis == 0:
            col_number = data.shape[1]
            if data.shape[0] * data.shape[1] > sample_number:
                row_number: int = sample_number // col_number

                for i in tqdm(range(col_number)):
                    _vector_ = down_sampling_data(data.X[:, i], row_number)
                    sns.kdeplot(np.array(_vector_).flatten(), shade=True, fill=True)
            else:
                for i in tqdm(range(col_number)):
                    sns.kdeplot(np.array(data.X[:, i]).flatten(), shade=True, fill=True)

            if is_legend:
                ax.legend(list(adata.var.index))

        elif axis == 1:
            row_number = data.shape[0]
            if data.shape[0] * data.shape[1] > sample_number:
                col_number: int = sample_number // row_number

                for i in tqdm(range(row_number)):
                    _vector_ = down_sampling_data(data.X[i, :], col_number)
                    sns.kdeplot(np.array(_vector_).flatten(), shade=True, fill=True)
            else:
                for i in tqdm(range(row_number)):
                    sns.kdeplot(np.array(data.X[i, :]).flatten(), shade=True, fill=True)

            if is_legend:
                ax.legend(list(adata.obs.index))

        plot_end(fig, output, show)



# -*- coding: UTF-8 -*-

from ._anndata_ import adata_map_df, adata_group
from ._scanpy_ import filter_data, get_difference_genes
from ._scvi_ import poisson_vi
from ._gsea_ import gsea_enrichr, get_gene_enrichment

from ._snapatac_ import (
    get_sc_atac,
    merge_sc_atac,
    get_gene_expression,
    get_peak_matrix,
    get_tf_data,
    get_difference_peaks
)

__all__ = [
    "poisson_vi",
    "gsea_enrichr",
    "get_gene_enrichment",
    "adata_map_df",
    "filter_data",
    "get_difference_genes",
    "adata_group",
    "get_sc_atac",
    "merge_sc_atac",
    "get_gene_expression",
    "get_peak_matrix",
    "get_tf_data",
    "get_difference_peaks"
]



# -*- coding: UTF-8 -*-

from typing import Optional, Tuple, Union

from pandas import DataFrame
from anndata import AnnData
from matplotlib import pyplot as plt
import matplotlib.dates as mdates
import seaborn as sns

from .. import util as ul
from ..preprocessing import adata_map_df
from ..util import path, plot_color_types, collection, plot_end

__name__: str = "plot_line"


def stability_line(
    data: Union[AnnData, DataFrame],
    x: str,
    y: str,
    layer: Optional[str] = None,
    width: float = 2,
    height: float = 2,
    title: Optional[str] = None,
    x_name: Optional[str] = None,
    y_name: Optional[str] = None,
    label: Optional[str] = None,
    legend: Optional[str] = None,
    legend_list: list = None,
    start_color_index: int = 0,
    color_step_size: int = 0,
    color_type: str = "set",
    colors: list = None,
    line_width: float = 1.5,
    x_name_rotation: float = 65,
    x_ticks: Optional[Union[int, collection]] = None,
    y_limit: Tuple[float, float] = (0, 1),
    output: Optional[path] = None,
    is_str: bool = True,
    show: bool = True
) -> None:
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        # noinspection DuplicatedCode
        fig, ax = plt.subplots(figsize=(width, height))
        new_data = data.copy()

        if isinstance(new_data, AnnData):

            if legend_list is not None:

                index_list = []
                label_list = list(new_data.var[label])

                for lab in range(len(label_list)):
                    if legend_list.count(label_list[lab]) > 0:
                        index_list.append(lab)

                if legend_list is not None:
                    new_data = new_data[:, index_list]

            # judge layers
            if layer is not None:

                if layer not in list(new_data.layers):
                    ul.log(__name__).error("The `layer` parameter needs to include in `adata.layers`")
                    raise ValueError("The `layer` parameter needs to include in `adata.layers`")

                new_data.X = new_data.layers[layer]

            # DataFrame
            ul.log(__name__).info(f"to DataFrame")
            df: DataFrame = adata_map_df(new_data, column="value")

        elif isinstance(new_data, DataFrame):

            if legend_list is not None:
                df: DataFrame = new_data[new_data[label].isin(legend_list)].copy()
            else:
                df: DataFrame = new_data.copy()

        else:
            ul.log(__name__).error(f"The `data` parameter only support `AnnData` and `DataFrame` class types.")
            raise ValueError(f"The `data` parameter only support `AnnData` and `DataFrame` class types.")

        if legend is None:
            legend = "category"

        df[legend] = df[label].copy()
        new_data_columns = list(df.columns)

        hue_types = list(set(df[legend]))

        # noinspection DuplicatedCode
        if colors is not None:
            palette = colors
        else:
            if "color" in new_data_columns:
                palette = df["color"]
            else:
                palette = []

                for i in range(len(hue_types)):
                    palette.append(plot_color_types[color_type][start_color_index + i * color_step_size + i])

        if x_name is not None:
            plt.xlabel(x_name)

        if y_name is not None:
            plt.ylabel(y_name, rotation=90)

        if title is not None:
            plt.title(title)

        # sns.set_theme(style="whitegrid")
        ax.set(ylim=y_limit)
        sns.despine()

        if is_str:
            df[x] = df[x].astype(str)

        chart = sns.lineplot(data=df, ax=ax, x=x, y=y, hue=legend, palette=palette, linewidth=line_width)

        if is_str:
            locator = mdates.DayLocator(interval=1)
            chart.xaxis.set_major_locator(locator)

            ax.set_xticklabels(ax.get_xticklabels(), rotation=x_name_rotation)
        else:
            plt.xticks(x_ticks, rotation=x_name_rotation)

        plot_end(fig, output, show)



# -*- coding: UTF-8 -*-

import matplotlib

from ._graph_ import graph, communities_graph, network_two_types
from ._heat_map_ import heatmap, heatmap_annotation
from ._scatter_ import scatter_base, scatter_atac, scatter_trait, volcano_base, manhattan_causal_variant
from ._violin_ import violin_base, violin_trait
from ._box_ import box_base, box_trait
from ._kde_ import kde
from ._line_ import stability_line
from ._bar_ import bar_trait, class_bar, bar, two_bar, bar_significance
from ._barcode_ import barcode_base, barcode_trait
from ._pie_ import pie_trait, pie_label, base_pie
from ._bubble_ import bubble
from ._radar_ import radar, radar_trait
from ._venn_ import three_venn, two_venn

from ._core_ import (
    trs_plot,
    group_heatmap,
    map_df_plot,
    rate_bar_plot,
    init_score_plot,
    cell_cell_plot,
    data_plot,
    complete_ratio,
    rate_circular_bar_plot
)

matplotlib.set_loglevel("error")

__all__ = [
    "graph",
    "communities_graph",
    "network_two_types",
    "heatmap",
    "volcano_base",
    "heatmap_annotation",
    "scatter_base",
    "scatter_atac",
    "scatter_trait",
    "violin_base",
    "violin_trait",
    "manhattan_causal_variant",
    "barcode_base",
    "barcode_trait",
    "box_base",
    "box_trait",
    "bar_significance",
    "bar",
    "pie_trait",
    "pie_label",
    "two_bar",
    "kde",
    "stability_line",
    "class_bar",
    "bar_trait",
    "bubble",
    "group_heatmap",
    "map_df_plot",
    "rate_bar_plot",
    "trs_plot",
    "init_score_plot",
    "cell_cell_plot",
    "data_plot",
    "complete_ratio",
    "rate_circular_bar_plot",
    "radar",
    "radar_trait",
    "three_venn",
    "two_venn"
]



# -*- coding: UTF-8 -*-

from matplotlib import pyplot as plt
from matplotlib_venn import venn3, venn3_circles, venn2, venn2_circles

from .. import util as ul
from ..util import path, collection, type_set_colors, plot_end

__name__: str = "plot_venn"


def three_venn(
    set1: collection,
    set2: collection,
    set3: collection,
    name1: str = "Set1",
    name2: str = "Set2",
    name3: str = "Set3",
    width: float = 2,
    height: float = 2,
    colors: list = None,
    title: str = None,
    output: path = None,
    show: bool = True
) -> None:
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        fig = plt.figure(figsize=(width, height))

        if title is not None:
            plt.title(title)

        if colors is None:
            colors = type_set_colors[:3]

        if len(colors) < 3:
            ul.log(__name__).info(f"The value of colors requires three elements.")
            raise ValueError(f"The value of colors requires three elements.")
        elif len(colors) > 3:
            colors = colors[:3]

        ax1 = fig.add_subplot()

        set1 = set(set1)
        set2 = set(set2)
        set3 = set(set3)

        subsets = (set1, set2, set3)

        venn3(subsets=subsets, set_labels=(name1, name2, name3), ax=ax1, set_colors=colors)

        # noinspection PyTypeChecker
        venn3_circles(subsets=subsets, linestyle='dashed', linewidth=1, color="grey", ax=ax1)

        ax1.legend(loc='upper right')

        ax1.axis('off')

        plot_end(fig, output, show)


def two_venn(
    set1: collection,
    set2: collection,
    name1: str = "Set1",
    name2: str = "Set2",
    width: float = 2,
    height: float = 2,
    colors: list = None,
    title: str = None,
    output: path = None,
    show: bool = True
) -> None:
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        fig = plt.figure(figsize=(width, height))

        if title is not None:
            plt.title(title)

        if colors is None:
            colors = type_set_colors[:2]

        if len(colors) < 2:
            ul.log(__name__).info(f"The value of colors requires three elements.")
            raise ValueError(f"The value of colors requires three elements.")
        elif len(colors) > 2:
            colors = colors[:2]

        ax1 = fig.add_subplot()

        set1 = set(set1)
        set2 = set(set2)

        venn2((set1, set2), set_labels=(name1, name2), ax=ax1, set_colors=colors)

        # noinspection PyTypeChecker
        venn2_circles(subsets=(set1, set2), linestyle='dashed', linewidth=1, color="grey", ax=ax1)

        ax1.legend(loc='upper right')

        ax1.axis('off')

        plot_end(fig, output, show)



# -*- coding: UTF-8 -*-

import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns
from pandas import DataFrame

from .. import util as ul
from ..util import path, plot_end

__name__: str = "plot_bubble"


def bubble(
    df: DataFrame,
    x: str,
    y: str,
    hue: str = None,
    size: str = None,
    width: float = 2,
    height: float = 2,
    title: str = None,
    output: path = None,
    show: bool = True
):
    if output is None and not show:
        ul.log(__name__).warning(f"At least one of the `output` and `show` parameters is required")
    else:
        fig, ax = plt.subplots(figsize=(width, height))

        if title is not None:
            plt.title(title)

        if size is not None:
            _size_ = df[size].values
            sizes = (np.array(_size_).min(), np.array(_size_).max())
        else:
            sizes = None

        sns.relplot(
            x=x,
            y=y,
            hue=hue,
            size=size,
            sizes=sizes,
            alpha=.5,
            palette="muted",
            height=6,
            data=df
        )

        plot_end(fig, output, show)
